import os
import re
import pandas as pd
import xursparks

class ItemReader:

    def excel_file_reader(self, file_path: str):
        # Dictionary to store dataframes by sheet name
        tagged_dataframes = {}
        
        # Iterate through all files in the given directory
        for file in os.listdir(file_path):
            if file.endswith('.xlsx') or file.endswith('.xls'):
                # Extract the filename without the extension
                filename = os.path.splitext(file)[0]
                
                # Construct full file path
                full_file_path = os.path.join(file_path, file)
                
                # Read the Excel file
                excel_file = pd.ExcelFile(full_file_path)
                
                # Iterate over all sheet names in the Excel file
                for sheet_name in excel_file.sheet_names:
                    # Read each sheet into a dataframe
                    sheet_df = pd.read_excel(full_file_path, sheet_name=sheet_name)
                    sheet_df = sheet_df.astype(str)       

                    # Store the dataframe in the dictionary with the key as (filename, sheet_name)
                    tagged_dataframes[(filename, sheet_name)] = sheet_df
        
        return tagged_dataframes
    
    # create function for specific file

    # create function that gets data given keyword, then 2nd keyword, then no. columns

    def json_file_reader(self, file_path: str, file_name: str):
        # Compile the regex pattern for matching file names
        regex_pattern = re.compile(file_name)
        
        # Search for files in the directory that match the pattern and have .json extension
        matching_files = [f for f in os.listdir(file_path) if f.endswith('.json') and regex_pattern.search(f)]
        
        # If no files match, return None
        if not matching_files:
            print("No matching JSON files found.")
            return None
        
        # If multiple files match, handle it accordingly (e.g., use the first match or process all)
        # Here, we'll assume the first match for simplicity
        json_file = matching_files[0]
        json_file_path = os.path.join(file_path, json_file)

        spark = xursparks.getSparkSession()
        
        sparkdf = spark.read.json(json_file_path)

        return sparkdf
    
    def sheet_name_extractor(self, file_path: str,
                             pattern: str):
        
        # Read the Excel file
        excel_file = pd.ExcelFile(file_path)

        # Get the list of sheet names
        sheet_names = excel_file.sheet_names

        # Loop through the sheet names to find the first match
        for sheet_name in sheet_names:
            if re.match(pattern.lower(), sheet_name.lower()):  # Make the comparison case-insensitive
                return sheet_name

        raise NoMatchingSheetError(f"No matching sheet found for pattern: {pattern}")
    
class NoMatchingSheetError(Exception):
    """Custom exception for no matching sheet found."""
    pass


import os
from dataclasses import dataclass
from itertools import islice

import openpyxl
from openpyxl.utils import get_column_letter, column_index_from_string, range_boundaries
from openpyxl.worksheet.cell_range import CellRange

import pandas as pd

from typing import Dict, List, Any, Tuple

@dataclass
class Keyword:
    keyword: str
    col: str = None
    row: str = None
    col_property: str = None
    row_property: str = None


class FileReader:
    """Class that helps extracts tables from an excel file"""
    def __init__(self, 
                 filename: str=None,
                 file_stream: List[list] = None, 
                 sheetname: List[str]|str = None,
                 column_num : int = None,
                 row_num: int = None,
                 table_start_column: str = None,
                 table_end_column: str = None,
                 table_start_row: str = None,
                 table_end_row: str = None,
                 keyword_start: str = None, 
                 keyword_end: str = None, 
                 keyword_start_column: str = None,
                 keyword_end_column: str = None,
                 keyword_start_row: str|int = None,
                 keyword_end_row: str|int = None,
                 keyword_start_property: str | Dict[str, str] = None,
                 keyword_end_property: str | Dict[str, str] = None,
                 pandas_parameters: Dict[str, Any] = None,
                 tolerance: int = 0,
                 unmerge_and_fill: bool = True,
                 **kwargs):
        """Initializes a FileReader Class.
        Input:
            filename: Path of the excel file
            file_stream: List of list holding values if from loading google sheet file
            sheetname: Name of the sheet/s to extract data from. Can be a list
            column_num: Number of Columns. 
            row_num: Number of Rows.
            table_start_column: Column in which the table starts.
                                If None, it is assumed to be the same column as with keyword_start.
            table_start_end: Row in which the table starts.
                             If None, it is assumed to be the same row as with keyword_start.
            
            keyword_start: Keyword (if any) to be used to find the start of the table
            keyword_end: Keyword (if any) to be used to find the end of the table
            keyword_start_column: Column Name/Header (if any) to be used to find start keyword.
                                - If None, searches whole sheet for the keyword.
            keyword_end_column: Column Name/Header (if any) to be used to find end keyword.
                                - If None, searches all available columns for the keyword.
            keyword_start_row: Row Number/Name (if any) to be used to find start keyword.
                                - If None, searches all available rows for the keyword.
                                            - If None, searches all available columns for the keyword.
            keyword_end_row: Row Number/Name (if any) to be used to find end keyword.
                                - If None, searches all available rows for the keyword.
            keyword_start_property: Accepts either None, 'after', 'before', and dict with keys 'row', and 'col'.
                                - If None, keyword indicates the start of the data table.
                                - If 'before', keyword indicates the start of the data table is row before keyword.
                                - If 'after', keyword indicates the start of the table is row after keyword.
                                - If dict is inputted, keys should be 'row' and 'col'. This is so that the parameters could
                                  be more specific on whether the before or after keyword means before or after the row or column of keyword.
            keyword_end_property: The same as keyword_start property except indicates the end of the table.
            tolerance: tolerance for blank rows/columns when searching for table
            unmerge_and_fill: If True, unmerges and fills all merged cells. Defaults to True
        """
        if filename != None and file_stream !=None:
            raise ValueError("Only use one of'filename' and 'file_stream' parameters")
        elif filename!=None and file_stream is None:
            self.filename = filename

        elif filename is None and file_stream != None:
            self.file_stream = file_stream
        self.sheetname = sheetname
        self.tolerance = tolerance
        
        self.table_dimensions = {'num_col': column_num,
                                 'num_row': row_num,
                                 'start_col': table_start_column,
                                 'start_row': table_start_row,
                                 'end_col': table_end_column,
                                 'end_row': table_end_row}

        if not isinstance(keyword_start_property, dict) and keyword_start_property not in [None, 'before', 'after']:
            raise ValueError("Invalid value for 'keyword_start_property'. Should be None, 'before' or 'after")
        elif isinstance(keyword_start_property, dict):
            for key, value in keyword_start_property.items():
                if key not in ['row', 'col']:
                    raise ValueError("Invalid value for keys in 'keyword_start_property'. Should be, 'row', 'col'")
                elif value not in [None, 'before', 'after']:
                    raise ValueError(f"Invalid value for 'keyword_start_property['{key}']'. Should be None, 'before' or 'after")
        
        if not isinstance(keyword_end_property, dict) and keyword_end_property not in [None, 'before', 'after','all']:
            raise ValueError("Invalid value for 'keyword_end_property'. Should be None, 'before' or 'after")
        elif isinstance(keyword_end_property, dict):
            for key, value in keyword_end_property.items():
                if key not in ['row', 'col']:
                    raise ValueError("Invalid value for keys in 'keyword_end_property'. Should be, 'row', 'col'")
                elif value not in [None, 'before', 'after', 'all']:
                    raise ValueError(f"Invalid value for 'keyword_end_property['{key}']'. Should be None, 'before' or 'after")
        
        def initialize_keyword(keyword, col, row, params):
            word = Keyword(keyword=keyword,
                          col=col,
                          row=row)
            if params is not None:
                if isinstance(params, dict):
                    for key, value in params.items():
                        if key =='col':
                            word.col_property = value
                        elif key == 'row':
                            word.row_property = value
                else:
                    word.row_property = params


            return word
        self.keyword = dict()
        self.keyword['start'] = initialize_keyword(keyword_start,
                                                    keyword_start_column,
                                                    keyword_start_row,
                                                    keyword_start_property)
            
        self.keyword['end'] = initialize_keyword(keyword_end,
                                                  keyword_end_column,
                                                  keyword_end_row,
                                                  keyword_end_property)
        
        self.pandas_parameters = {} if pandas_parameters is None else pandas_parameters
        self.unmerge_and_fill = unmerge_and_fill
        self.__load_file__()

    def __load_file__(self):
        if self.filename:
            self.wb = openpyxl.load_workbook(self.filename, data_only=True)
            self.sheet = self.wb[self.sheetname]
        elif self.file_stream:
            self.wb = openpyxl.Workbook()
            self.sheet = self.wb.active
            self.sheet.title = self.sheetname
            for row in self.file_stream:
                self.sheet.append(row)

    def __find_loc__(self, position:str,property:str, cell:openpyxl.cell.cell.Cell, dimension: str):
        if dimension == 'col':
            if self.table_dimensions[f'{position}_col'] is None:
                if property == 'before':
                    loc = get_column_letter(cell.offset(column=-1).column)
                elif property == 'after':
                    loc = get_column_letter(cell.offset(column=+1).column)
                elif property is None:
                    loc = get_column_letter(cell.column)
            else:
                loc = self.table_dimensions[f'{position}_col']

        elif dimension =='row':
            if self.table_dimensions[f'{position}_row'] is None:
                if property == 'before':
                    loc = cell.row - 1
                elif property == 'after':
                    loc = cell.row + 1
                else:
                    loc = cell.row
            else:
                loc = self.table_dimensions['start_row']

        else:
            raise ValueError(f"'dimension' must be either 'row', or 'col'. Not {dimension}")
        return loc
    
    def __find_loc_all__(self, cell:openpyxl.cell.cell.Cell, start_location: openpyxl.cell.cell.Cell, dimension:str, opp_param:str=None, start:bool=False)->openpyxl.cell.cell.Cell:
        """if start is True, then tries to find the end of the table without any end condition, except tolerance starting from start location"""

        params = {'min_col': self.sheet[start_location[0]][0].column,
                  'min_row': start_location[1]
                } 
        
        if not start:
            if dimension == 'col':
                key = 'max_row'
            elif dimension == 'row':
                key = 'max_col'

            if opp_param == 'before':
                params[key] = cell.row - 1
            elif opp_param is None:
                params[key] = cell.row
            elif opp_param == 'after':
                params[key] = cell.row + 1

        current_cell = None
        iterator = self.sheet.iter_cols if dimension == 'col' else self.sheet.iter_rows
        consecutive_blank_count = 0

        for row in iterator(**params):
            if all(cell.value is None for cell in row):
                consecutive_blank_count += 1
                if consecutive_blank_count > self.tolerance:
                    break
            else:
                consecutive_blank_count = 0
                current_cell = row[-1]


        """current_cell = None
        iterator = self.sheet.iter_cols if dimension == 'col' else self.sheet.iter_rows
        for row in iterator(**params):
            if all(cell.value is None for cell in row):
                break
            for cell in row:
                current_cell = cell"""

        return get_column_letter(current_cell.column), current_cell.row

    def __get_location__(self, position:str, cell:openpyxl.cell.cell.Cell, start_location:openpyxl.cell.cell.Cell=None)->openpyxl.cell.cell.Cell:
        ATTRIBUTES = ['col', 'row']
        for key in ATTRIBUTES:
            value = getattr(self.keyword[position], f"{key}_property")
            if position == 'end' and value == 'all':
                col, row = self.__find_loc_all__(cell, start_location,key, self.keyword[position].col_property if key == 'row' else self.keyword[position].row_property)

            elif key == 'col':
                col = self.__find_loc__(position, value, cell,key)
            elif key == 'row':
                row = self.__find_loc__(position, value, cell, key )

        cell_pos = str(col)+str(row)

        return self.sheet[cell_pos]

    def __find_table_position__(self, position:str, start_location:tuple=None)->tuple:
        """Finds start or end location of table."""
        def iterate_cells(keyword:Any, sheet: openpyxl.worksheet.worksheet.Worksheet, dimension:str, col:str=None,params:dict=None):
            params = dict() if params is None else params
            if position == 'end':
                params['min_col'] = column_index_from_string(start_location[0])
                params['min_row'] = start_location[1]

            if dimension == 'rows':
                for row in sheet.iter_rows(**params):
                    for cell in row:
                        if cell.value == keyword:
                            return cell
                        
            elif dimension == 'cols' and col is not None:
                for cell in sheet[col]:
                    if cell.value == keyword:
                        return cell
                    
            elif dimension == 'cols':
                for col in sheet.iter_cols(**params):
                    for cell in col:
                        if cell.value == keyword:
                            return cell 
            else:
                return None

        if position not in ['start', 'end']:
            return ValueError(f"'position' parameter should either be 'start' or 'end'. Not {position}")

        position_cell = None
        if self.keyword[position].col is not None:
            if self.keyword[position].row is not None:
                if position=='end':
                    params={
                        'min_row': start_location[1]+1,
                        'min_col': self.sheet[self.keyword[position].col][0].column,
                        'max_col': self.sheet[self.keyword[position].col][0].column
                    }
                    position_cell = iterate_cells(keyword=self.keyword[position].keyword,sheet= self.sheet, dimension='cols', params=params)
                else:
                    position_cell = iterate_cells(keyword=self.keyword[position].keyword,sheet= self.sheet, dimension='cols', col= self.keyword[position].col)
                
                if position_cell is None:
                    raise ValueError(f"Keyword {self.keyword[position].keyword} not found in {self.keyword[position].col}.")
                else:
                    cellname = str(self.keyword[position].col) + str(self.keyword[position].row)
                    position_cell = self.sheet[cellname]
            else: #if keyword[position].row is None
                position_cell = iterate_cells(keyword=self.keyword[position].keyword,
                                              sheet=self.sheet,
                                              dimension='cols',
                                              col=self.keyword[position].col)

        elif self.keyword[position].row is not None:
            params = {
                'min_row': int(self.keyword[position].row),
                'max_row': int(self.keyword[position].row)
            }
            position_cell = iterate_cells(keyword=self.keyword[position].keyword, sheet=self.sheet, dimension='cols', params=params)

        else: #iterates through all cells if no keyword location is specified
            position_cell = iterate_cells(keyword=self.keyword[position].keyword, sheet=self.sheet, dimension='cols')


        location = self.__get_location__(position, position_cell, start_location)
        return (get_column_letter(location.column), location.row)

    def __find_table_range__(self) -> Tuple[Tuple[int,int], Tuple[int,int]]:
        
        if self.keyword['start'].keyword is not None:
            start_location = self.__find_table_position__('start')

        if self.keyword['end'].keyword is not None:
            end_location = self.__find_table_position__('end', start_location = start_location)
        else:
            end_location = self.__find_loc_all__(self.sheet[start_location[0]+str(start_location[1])],
                                                 start_location=start_location,
                                                 dimension='col',
                                                 start=True)

            
        return start_location, end_location

    def get_merged_cell_ranges(self)->List:
        merged_cells = []
        for merged_cell in self.sheet.merged_cells.ranges:
            merged_cells.append(merged_cell)

        return merged_cells
        
    def get_merged_cell_ranges_in_table_range(self, merged_cells_range:List[CellRange], table_range:str):
        table_range = CellRange(table_range)
        filtered_merged_cells = []
        for merged_cell in merged_cells_range:
            if table_range.issuperset(merged_cell):
                filtered_merged_cells.append(merged_cell)

        return filtered_merged_cells
    
    def unmerge_and_fill_cell(self,cell_range:str):
        # Get the top-left cell of the merged range
        min_col, min_row, max_col, max_row = range_boundaries(cell_range)
        top_left_cell = self.sheet.cell(min_row, min_col)
        value = top_left_cell.value

        # Unmerge the cells
        self.sheet.unmerge_cells(cell_range)

        # Fill all cells in the previously merged range with the value
        for row in range(min_row, max_row + 1):
            for col in range(min_col, max_col + 1):
                self.sheet.cell(row, col).value = value

    def get_table(self) -> pd.DataFrame:
        if self.keyword['start'].keyword is None and self.keyword['end'].keyword is None:
            ### ingests whole sheet
            if self.unmerge_and_fill:
                ranges = self.get_merged_cell_ranges()
                for cell_range in ranges:
                    self.unmerge_and_fill_cell(cell_range.coord)
            
            data = self.sheet.values
            cols = next(data)[1:]
            data = list(data)
            idx = [r[0] for r in data]
            data = (islice(r, 1, None) for r in data)
            df = pd.DataFrame(data, index=idx, columns=cols, **self.pandas_parameters)
        else:
            ### ingests only range
            start_location, end_location = self.__find_table_range__()
            table_range = f"{start_location[0]}{start_location[1]}:{end_location[0]}{end_location[1]}"

            if self.unmerge_and_fill:
                ranges = self.get_merged_cell_ranges()
                ranges = self.get_merged_cell_ranges_in_table_range(ranges, table_range)
                for cell_range in ranges:
                    self.unmerge_and_fill_cell(cell_range.coord)

            cell_range = self.sheet[table_range]
            values_list = []
            for row in cell_range:
                row_values = [cell.value for cell in row]
                values_list.append(row_values)
            df = pd.DataFrame(values_list[1:], columns=values_list[0], **self.pandas_parameters)
            
        """
        skiprows = start_location[1] -1# need to compensate for the one row to be the header and for index-0
        nrows = end_location[1] - start_location[1] + 1 # need to compensate for index-0
        usecols = f"{start_location[0]}:{end_location[0]}"
        df = pd.read_excel(self.filename, self.sheetname, skiprows=skiprows, nrows=nrows, usecols=usecols, **self.pandas_parameters)"""
        
        return df
    
    def get_keywords(self, keyword:str, row:str=None, col:str=None, first_instance_only=True)->List[Keyword]:
        """"Gets keywords if there are multiple in one sheet"""
        keywords = []

        params = {}
        if row is not None:
            params['min_row'] = row
            params['max_row'] = row
        
        if col is not None:
            params['min_col'] = column_index_from_string(col)
            params['max_col'] = column_index_from_string(col)

        first_instance_only = True

        if first_instance_only:
            existing_keywords = set()

        for row in self.sheet.iter_cols(**params):
            for cell in row:
                cell_value = str(cell.value)
                if keyword is not None and keyword in cell_value:
                    if not first_instance_only or (first_instance_only and cell.value not in existing_keywords):
                        keywords.append(Keyword(
                            keyword=cell.value,
                            col=cell.column_letter,
                            row=cell.row
                        ))
                        if first_instance_only:
                            existing_keywords.add(cell.value)
                    
                    
        return keywords
    



from pydantic import BaseModel, Field
from langchain.tools import BaseTool
from typing import Type
from langchain.schema import SystemMessage
from langchain_community.chat_models import ChatOpenAI
from langchain.agents import initialize_agent
from langchain.agents import AgentType
from datetime import date
import pandas as pd
import json
from sqlalchemy import create_engine
from typing import Optional

class FinanceAgent:

    def __init__(self, temperature:float = None,
                llm_ver:str = None,
                model:str = None,
                embed_batch_size:int = None,
                streaming:bool = None,
                verbose:bool = None,
                path:str = None,
                db_domain:str = None,
                db_port:str = None,
                db_name:str = None,
                db_password:str = None) -> None:
        self.temperature = temperature
        self.llm_ver = llm_ver
        self.model = model
        self.embed_batch_size = embed_batch_size
        self.streaming = streaming
        self.verbose = verbose
        self.path = path
        self.db_domain = db_domain
        self.db_port = db_port
        self.db_name = db_name
        self.db_password = db_password

    def get_agent(self, doc:list[any],
                  content:str = None):
        year = date.today().year
        print(f'GET AGENT YEAR={year}')
        companies = []
        companies = [x.name for x in doc]
        print(companies)
        system_message = SystemMessage(content)
        accounting_journal_tool = AccountingJournalTool(
            path = self.path,
            db_domain=self.db_domain,
            db_port=self.db_port,
            db_name=self.db_name,
            db_password=self.db_password
        )
        tools = [
            accounting_journal_tool
        ]
        agent_kwargs = {
            "system_message": system_message,
        }
        llm = ChatOpenAI(temperature=self.temperature, model=self.model, verbose=self.verbose)
        agent = initialize_agent(
            tools,
            llm,
            agent=AgentType.OPENAI_FUNCTIONS,
            verbose=self.verbose,
            agent_kwargs=agent_kwargs,
        )

        return agent

class AccountingJournalInput(BaseModel):
    """Inputs for get_content function"""
    company: str = Field(
        description="The company that you want to get the journal entries for.")
    start_date: date = Field(
        description="The start date of the journal entries that you want to query")
    end_date: date = Field(
        description="The end date of the journal entries that you want to query")


class AccountingJournalTool(BaseTool):
    name = "accounting_journal_tool"
    description = "Use to get a pandas dataframe of a company's accounting journal entries"
    args_schema: Type[BaseModel] = AccountingJournalInput
    
    # Make the database fields optional with a default of None
    path: Optional[str] = Field(None, description="The config path")
    db_domain: Optional[str] = Field(None, description="The database domain")
    db_port: Optional[str] = Field("3306", description="The database port")
    db_name: Optional[str] = Field(None, description="The database name")
    db_password: Optional[str] = Field(None, description="The database password")



    def get_data(self, company, start_date=None, end_date=None):

        print(f'{company} {start_date} {end_date}')
        conn_string = self.get_conn_string()
        if start_date == None:
            start_date = date(date.today().year, 1, 1)
            end_date = date(date.today().year, 12, 31)
        # print(start_date)
        # print(end_date)
        sql = f'SELECT * FROM `tabJournal Entry` WHERE company="{company}" AND posting_date BETWEEN "{start_date}" AND "{end_date}" ORDER BY posting_date'
        # print (sql)
        db_connection = create_engine(conn_string)
        df = pd.read_sql(sql, con=db_connection)
        print(df.head())
        return df

    def get_conn_string(self):

        print("self.path: ", self.path)
        print("self.db_domain: ", self.db_domain)
        print("self.db_port: ", self.db_port)
        path = f'{self.path}/site_config.json'
        db_domain = self.db_domain
        db_port = self.db_port
        with open(path) as file:
            j = json.load(file)
            #print(j)

            conn_string = f'mariadb+pymysql://{j["db_name"]}:{j["db_password"]}@{db_domain}:{db_port}/{j["db_name"]}'

        #print("conn_string:  ", conn_string)
        return conn_string

    def _run(self, company: str, start_date: date, end_date: date):
        return self.get_data(company, start_date, end_date)

    def _arun(self, cpmpany: str, attr: str):
        raise NotImplementedError("error here")


import logging
from datetime import datetime
import os

class Logger:
    """
    A class to handle logging functionality.
    """
    def __init__(self, log_dir: str = "logs"):
        """
        Initialize logger with log directory.
        
        Args:
            log_dir (str): Directory to store log files
        """
        self.log_dir = log_dir
        
        # Create logs directory if it doesn't exist
        if not os.path.exists(log_dir):
            os.makedirs(log_dir)
            
        # Configure logging
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(logging.INFO)
        
        # Create formatter
        formatter = logging.Formatter(
            '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        )
        
        # Create handlers
        # File handler - logs to file with date
        log_file = os.path.join(
            log_dir, 
            f"log_{datetime.now().strftime('%Y%m%d')}.log"
        )
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(formatter)
        
        # Stream handler - logs to console
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(formatter)
        
        # Add handlers to logger
        self.logger.addHandler(file_handler)
        self.logger.addHandler(console_handler)

    def log(self, message: str, level: str = "info"):
        """
        Log a message with specified level.
        
        Args:
            message (str): Message to log
            level (str): Logging level (debug, info, warning, error, critical)
        """
        level = level.lower()
        if level == "debug":
            self.logger.debug(message)
        elif level == "info":
            self.logger.info(message)
        elif level == "warning":
            self.logger.warning(message)
        elif level == "error":
            self.logger.error(message)
        elif level == "critical":
            self.logger.critical(message)
        else:
            self.logger.info(f"Unknown log level '{level}', using INFO: {message}")

    def __del__(self):
        """Cleanup handlers on object destruction"""
        for handler in self.logger.handlers[:]:
            handler.close()
            self.logger.removeHandler(handler)


import requests
import json

class Xurl:
    """
    A class to represent a cURL command.
    """
    def __init__(self, url: str,
                 headers: dict = None,
                 data: str = None,
                 params: dict = None,
                 json: dict = None,
                 auth: tuple = None,
                 ntlm_auth: object = None,
                 verify_ssl: bool = False):
        self.url = url
        self.headers = headers
        self.data = data
        self.params = params
        self.json = json
        self.auth = auth
        self.ntlm_auth = ntlm_auth  # Added to handle NTLM authentication   
        self.verify_ssl = verify_ssl  # Added to handle SSL verification

    #add code that will implement get and post methods using requests library
    def get(self):
        """
        Perform a GET request using the requests library.
        
        Returns:
            dict: JSON response with cookies
            
        Raises:
            requests.RequestException: If the request fails
        """
        try:
            # Build kwargs dictionary with only non-None values
            kwargs = {
                'url': self.url,
                'verify': self.verify_ssl
            }
            
            # Add optional parameters if they exist
            if self.headers:
                kwargs['headers'] = self.headers
            if self.params:
                kwargs['params'] = self.params
            if self.auth:
                kwargs['auth'] = self.auth
            if self.ntlm_auth:
                kwargs['auth'] = self.ntlm_auth

            # print all kwargs for debugging
            print(f"[xursparks.http.main][GET] Request parameters: {kwargs}")
            
            # Make the request with only non-None parameters
            response = requests.get(**kwargs)
            response.raise_for_status()
            
            # Get cookies from response
            cookies = response.cookies.get_dict()
            
            # Combine response JSON and cookies
            # Try to parse JSON, if fails return text response
            try:
                resp_data = response.json()
                resp_data['cookies'] = cookies
            except json.JSONDecodeError:
                resp_data = {
                    'text': response.text,
                    'cookies': cookies
                }
            # print(f"[xursparks.http.main][GET] Response data: {resp_data}")
            return resp_data
        except requests.RequestException as e:
            print(f"Request failed: {str(e)}")
            raise
    
    
    #add code that will implement post method using requests library
    def post(self):
        """
        Perform a POST request using requests library.
        
        Returns:
            dict: JSON response with cookies
            
        Raises:
            requests.RequestException: If the request fails
        """
        # print all kwargs for debugging
        print(f"[xursparks.http.main][POST] Request parameters: {kwargs}")
        try:
            # Build kwargs dictionary with only non-None values
            kwargs = {
                'url': self.url,
                'verify': self.verify_ssl
            }
            
            # Add optional parameters if they exist
            if self.headers:
                kwargs['headers'] = self.headers
            if self.params:
                kwargs['params'] = self.params
            if self.data:
                kwargs['data'] = self.data
            if self.json:
                kwargs['json'] = self.json
            if self.auth:
                kwargs['auth'] = self.auth
            
            # Make the request with only non-None parameters
            response = requests.post(**kwargs)
            response.raise_for_status()
            
            # Get cookies from response
            cookies = response.cookies.get_dict()
            
            # Combine response JSON and cookies
            try:
                resp_data = response.json()
                resp_data['cookies'] = cookies
            except json.JSONDecodeError:
                resp_data = {
                    'text': response.text,
                    'cookies': cookies
                }
            # print(f"[xursparks.http.main][POST] Response data: {resp_data}")
            return resp_data
        except requests.RequestException as e:
            print(f"Request failed: {str(e)}")
            raise

if __name__ == "__main__":
    pass



import os
import hashlib
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes # Keep modes for AES-GCM
from cryptography.hazmat.primitives.ciphers import aead # Corrected import for ChaCha20Poly1305
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa, padding # Corrected 'hazhat' to 'hazmat'
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives import serialization # Import for key serialization
from cryptography.exceptions import InvalidTag, UnsupportedAlgorithm
from typing import Tuple, Union

class CipherSuite:
    """
    A Python class providing common encryption functionalities for strings
    using AES-GCM, ChaCha20-Poly1305, RSA-OAEP, and data hashing.

    Note: This class handles string to bytes conversion internally (UTF-8).
    All encryption functions return bytes, and decryption functions expect bytes
    and return a string.
    """

    def __init__(self):
        """
        Initializes the CipherSuite. The backend is implicitly loaded when
        cryptographic operations are performed, preventing pickling issues.
        """
        # We explicitly don't store default_backend() here as an instance variable
        # because it causes pickling issues with Spark UDFs.
        # Instead, we'll get the backend when needed within the methods.
        pass

    def _get_backend(self):
        """Helper to get the cryptography backend."""
        return default_backend()

    def _string_to_bytes(self, text: str) -> bytes:
        """Helper to encode a string to bytes."""
        return text.encode('utf-8')

    def _bytes_to_string(self, data: bytes) -> str:
        """Helper to decode bytes to a string."""
        return data.decode('utf-8')

    # --- Data Hashing ---

    def compute_hash(self, data_string: str, algorithm: str = 'sha256') -> str:
        """
        Computes the cryptographic hash of a string.

        Args:
            data_string (str): The string to hash.
            algorithm (str): The hashing algorithm to use (e.g., 'sha256', 'sha512', 'md5').
                             Note: MD5 is considered insecure for cryptographic purposes.

        Returns:
            str: The hexadecimal representation of the hash.

        Raises:
            ValueError: If an unsupported hashing algorithm is specified.
        """
        data_bytes = self._string_to_bytes(data_string)
        hasher = None
        if algorithm == 'sha256':
            hasher = hashlib.sha256()
        elif algorithm == 'sha512':
            hasher = hashlib.sha512()
        elif algorithm == 'md5':
            # MD5 is provided for completeness but not recommended for security-critical uses
            print("Warning: MD5 is not recommended for cryptographic security.")
            hasher = hashlib.md5()
        else:
            raise ValueError(f"Unsupported hashing algorithm: {algorithm}. Choose from 'sha256', 'sha512', 'md5'.")

        hasher.update(data_bytes)
        return hasher.hexdigest()

    # --- AES-GCM (Symmetric Encryption) ---

    def generate_aes_key(self, key_size_bits: int = 256) -> bytes:
        """
        Generates a new AES symmetric key.
        AES supports 128, 192, or 256-bit keys.

        Args:
            key_size_bits (int): The desired key size in bits (128, 192, or 256).

        Returns:
            bytes: The randomly generated AES key.

        Raises:
            ValueError: If an unsupported key size is requested.
        """
        if key_size_bits not in [128, 192, 256]:
            raise ValueError("AES key size must be 128, 192, or 256 bits.")
        return os.urandom(key_size_bits // 8) # Convert bits to bytes

    def encrypt_aes(self, plaintext_string: str, key: bytes) -> Tuple[bytes, bytes, bytes]:
        """
        Encrypts a string using AES in GCM (Galois/Counter Mode).
        GCM provides both confidentiality and integrity (authentication).

        Args:
            plaintext_string (str): The string to encrypt.
            key (bytes): The AES key (16, 24, or 32 bytes).

        Returns:
            Tuple[bytes, bytes, bytes]: A tuple containing (nonce, ciphertext, tag).
                                       All three are required for successful decryption.
        """
        plaintext_bytes = self._string_to_bytes(plaintext_string)
        # Nonce (Initialization Vector) must be unique for each encryption with the same key
        # For GCM, 12 bytes (96 bits) is the recommended nonce length.
        nonce = os.urandom(12)

        # Get backend locally within the method
        cipher = Cipher(algorithms.AES(key), modes.GCM(nonce), backend=self._get_backend())
        encryptor = cipher.encryptor()

        # Encrypt the data
        ciphertext = encryptor.update(plaintext_bytes) + encryptor.finalize()
        tag = encryptor.tag # Get the authentication tag

        return nonce, ciphertext, tag

    def decrypt_aes(self, nonce: bytes, ciphertext: bytes, tag: bytes, key: bytes) -> Union[str, None]:
        """
        Decrypts AES-GCM encrypted bytes back to a string.

        Args:
            nonce (bytes): The nonce used during encryption.
            ciphertext (bytes): The encrypted data.
            tag (bytes): The authentication tag.
            key (bytes): The AES key.

        Returns:
            Union[str, None]: The decrypted string, or None if decryption/authentication fails.
        """
        try:
            # Get backend locally within the method
            cipher = Cipher(algorithms.AES(key), modes.GCM(nonce, tag), backend=self._get_backend())
            decryptor = cipher.decryptor()

            # Decrypt the data and verify the tag
            plaintext_bytes = decryptor.update(ciphertext) + decryptor.finalize()
            return self._bytes_to_string(plaintext_bytes)
        except InvalidTag:
            print("AES Decryption failed: Authentication tag is invalid. Data may be tampered with or key is incorrect.")
            return None
        except Exception as e:
            print(f"AES Decryption failed: {e}")
            return None

    # --- ChaCha20-Poly1305 (Symmetric Stream Cipher with Authentication) ---

    def generate_chacha20_key(self) -> bytes:
        """
        Generates a new ChaCha20 symmetric key (32 bytes / 256 bits).

        Returns:
            bytes: The randomly generated ChaCha20 key.
        """
        return os.urandom(32)

    def encrypt_chacha20(self, plaintext_string: str, key: bytes) -> Tuple[bytes, bytes, bytes]:
        """
        Encrypts a string using ChaCha20-Poly1305.
        This provides both confidentiality and integrity.

        Args:
            plaintext_string (str): The string to encrypt.
            key (bytes): The ChaCha20 key (32 bytes).

        Returns:
            Tuple[bytes, bytes, bytes]: A tuple containing (nonce, ciphertext, tag).
                                       All three are required for successful decryption.
        """
        plaintext_bytes = self._string_to_bytes(plaintext_string)
        # Nonce for ChaCha20-Poly1305 is usually 12 bytes (96 bits).
        nonce = os.urandom(12) 

        # Instantiate ChaCha20Poly1305 directly from aead module
        chacha_cipher = aead.ChaCha20Poly1305(key)
        
        # The encrypt method of ChaCha20Poly1305 returns ciphertext concatenated with the tag
        # associated_data is None for simple string encryption
        encrypted_with_tag = chacha_cipher.encrypt(nonce, plaintext_bytes, None)

        # Separate ciphertext and tag. Poly1305 tag is always 16 bytes.
        ciphertext = encrypted_with_tag[:-16]
        tag = encrypted_with_tag[-16:]

        return nonce, ciphertext, tag


    def decrypt_chacha20(self, nonce: bytes, ciphertext: bytes, tag: bytes, key: bytes) -> Union[str, None]:
        """
        Decrypts ChaCha20-Poly1305 encrypted bytes back to a string.

        Args:
            nonce (bytes): The nonce used during encryption.
            ciphertext (bytes): The encrypted data.
            tag (bytes): The authentication tag.
            key (bytes): The ChaCha20 key.

        Returns:
            Union[str, None]: The decrypted string, or None if decryption/authentication fails.
        """
        try:
            # Instantiate ChaCha20Poly1305 directly from aead module
            chacha_cipher = aead.ChaCha20Poly1305(key)

            # The decrypt method of ChaCha20Poly1305 expects ciphertext concatenated with the tag
            # associated_data is None for simple string encryption
            plaintext_bytes = chacha_cipher.decrypt(nonce, ciphertext + tag, None)
            return self._bytes_to_string(plaintext_bytes)
        except InvalidTag:
            print("ChaCha20-Poly1305 Decryption failed: Authentication tag is invalid. Data may be tampered with or key is incorrect.")
            return None
        except Exception as e:
            print(f"ChaCha20-Poly1305 Decryption failed: {e}")
            return None

    # --- RSA (Asymmetric Encryption) ---

    def generate_rsa_key_pair(self, key_size: int = 2048) -> Tuple[bytes, bytes]:
        """
        Generates a new RSA public/private key pair and returns them in PEM format.
        Recommended key size is 2048 bits or higher.

        Args:
            key_size (int): The desired key size in bits.

        Returns:
            Tuple[bytes, bytes]: A tuple containing (private_key_pem, public_key_pem).
                                 These are byte strings in PEM format.
        """
        private_key = rsa.generate_private_key(
            public_exponent=65537, # Common public exponent
            key_size=key_size,
            backend=self._get_backend() # Get backend locally for key generation
        )
        public_key = private_key.public_key()

        # Serialize keys to PEM format (bytes)
        private_pem = private_key.private_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PrivateFormat.PKCS8,
            # No encryption for simplicity in this example, but recommended for production
            encryption_algorithm=serialization.NoEncryption()
        )
        public_pem = public_key.public_bytes(
            encoding=serialization.Encoding.PEM,
            format=serialization.PublicFormat.SubjectPublicKeyInfo
        )
        return private_pem, public_pem

    def _load_rsa_public_key(self, public_key_pem: bytes) -> rsa.RSAPublicKey:
        """Helper to load an RSA public key from PEM bytes."""
        return serialization.load_pem_public_key(public_key_pem, backend=self._get_backend())

    def _load_rsa_private_key(self, private_key_pem: bytes) -> rsa.RSAPrivateKey:
        """Helper to load an RSA private key from PEM bytes."""
        return serialization.load_pem_private_key(private_key_pem, password=None, backend=self._get_backend()) # No password since we exported without one

    def encrypt_rsa(self, plaintext_string: str, public_key_pem: bytes) -> bytes:
        """
        Encrypts a string using RSA with OAEP padding.
        The public key is expected in PEM byte format.

        Args:
            plaintext_string (str): The string to encrypt.
            public_key_pem (bytes): The RSA public key in PEM byte format.

        Returns:
            bytes: The RSA-encrypted ciphertext.

        Raises:
            ValueError: If the plaintext is too long for the given RSA key.
        """
        public_key = self._load_rsa_public_key(public_key_pem)
        plaintext_bytes = self._string_to_bytes(plaintext_string)
        try:
            ciphertext = public_key.encrypt(
                plaintext_bytes,
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None
                )
            )
            return ciphertext
        except ValueError as e:
            # Re-raise the error after printing, as this is a critical input constraint
            raise ValueError(f"RSA Encryption failed: {e}. Plaintext too long for key size or padding.")
        except Exception as e:
            print(f"RSA Encryption failed: {e}")
            raise


    def decrypt_rsa(self, ciphertext: bytes, private_key_pem: bytes) -> Union[str, None]:
        """
        Decrypts RSA-encrypted bytes back to a string using OAEP padding.
        The private key is expected in PEM byte format.

        Args:
            ciphertext (bytes): The RSA-encrypted ciphertext.
            private_key_pem (bytes): The RSA private key in PEM byte format.

        Returns:
            Union[str, None]: The decrypted string, or None if decryption fails.
        """
        private_key = self._load_rsa_private_key(private_key_pem)
        try:
            plaintext_bytes = private_key.decrypt(
                ciphertext,
                padding.OAEP(
                    mgf=padding.MGF1(algorithm=hashes.SHA256()),
                    algorithm=hashes.SHA256(),
                    label=None
                )
            )
            return self._bytes_to_string(plaintext_bytes)
        except ValueError as e:
            print(f"RSA Decryption failed (bad decryption or padding): {e}")
            return None
        except Exception as e:
            print(f"RSA Decryption failed: {e}")
            return None

# --- Example Usage with Spark DataFrames ---
if __name__ == "__main__":
    # Import Spark-related components only if in the main execution block
    # This prevents import errors if the class is used without Spark
    try:
        from pyspark.sql import SparkSession
        from pyspark.sql.functions import udf, lit
        from pyspark.sql.types import StringType, BinaryType, StructType, StructField
        from pyspark.sql.types import ArrayType, ByteType # For potential tuple-like returns as array
    except ImportError:
        SparkSession = None
        udf = None
        lit = None
        StringType = None
        BinaryType = None
        StructType = None
        StructField = None
        ArrayType = None
        ByteType = None

    if SparkSession is None:
        print("PySpark is not installed. Skipping Spark DataFrame examples.")
        print("Please install it using: pip install pyspark")
    else:
        print("--- Initializing Spark Session ---")
        spark = SparkSession.builder \
            .appName("EncryptionDemo") \
            .getOrCreate()

        # Instantiate CipherSuite for key generation and initial setup on the driver
        # For UDFs, a new instance will be created on the worker
        driver_cipher_suite = CipherSuite()

        # --- Sample Data ---
        data = [
            (1, "123 Main St, Anytown, USA", "user1@example.com"),
            (2, "456 Oak Ave, Otherville, CA", "user2@example.com"),
            (3, "789 Pine Ln, Somewhere, TX", "user3@example.com"),
            (4, "101 Maple Rd, Nowhere, FL", "user4@example.com")
        ]
        columns = ["id", "address", "email"]
        df = spark.createDataFrame(data, columns)

        print("\n--- Original DataFrame ---")
        df.show(truncate=False)

        # --- Data Hashing Example (address column) ---
        print("\n--- Hashing 'address' column with SHA256 ---")
        # Instantiate CipherSuite within the UDF lambda to avoid pickling FFI object
        hash_udf = udf(lambda addr: CipherSuite().compute_hash(addr, 'sha256'), StringType())
        df_hashed = df.withColumn("address_hashed", hash_udf(df["address"]))
        df_hashed.show(truncate=False)

        # --- AES-GCM Encryption/Decryption (address column) ---
        print("\n--- AES-GCM Encryption/Decryption of 'address' column ---")
        # Generate key on driver and broadcast it
        aes_key = spark.sparkContext.broadcast(driver_cipher_suite.generate_aes_key())

        # Define UDF for AES encryption - CipherSuite instantiated within lambda
        aes_encrypt_udf = udf(lambda addr: CipherSuite().encrypt_aes(addr, aes_key.value),
                               StructType([
                                   StructField("nonce", BinaryType(), True),
                                   StructField("ciphertext", BinaryType(), True),
                                   StructField("tag", BinaryType(), True)
                               ]))

        df_aes_encrypted = df.withColumn("address_aes_encrypted", aes_encrypt_udf(df["address"]))
        print("DataFrame with AES Encrypted Address:")
        df_aes_encrypted.select("id", "address_aes_encrypted").show(truncate=False)

        # Define UDF for AES decryption - CipherSuite instantiated within lambda
        # Access elements by index [0] for nonce, [1] for ciphertext, [2] for tag
        aes_decrypt_udf = udf(lambda encrypted_tuple: \
                                CipherSuite().decrypt_aes(
                                    encrypted_tuple[0],  # nonce
                                    encrypted_tuple[1],  # ciphertext
                                    encrypted_tuple[2],  # tag
                                    aes_key.value
                                ), StringType())

        df_aes_decrypted = df_aes_encrypted.withColumn("address_aes_decrypted", aes_decrypt_udf(df_aes_encrypted["address_aes_encrypted"]))
        print("DataFrame with AES Decrypted Address (and original for comparison):")
        df_aes_decrypted.select("id", "address", "address_aes_decrypted").show(truncate=False)

        # Verify decryption
        df_aes_decrypted.createOrReplaceTempView("aes_data")
        match_count = spark.sql("SELECT COUNT(*) FROM aes_data WHERE address = address_aes_decrypted").collect()[0][0]
        total_count = df_aes_decrypted.count()
        print(f"AES Decryption Verification: {match_count}/{total_count} rows match original. {'SUCCESS' if match_count == total_count else 'FAILURE'}\n")

        # --- ChaCha20-Poly1305 Encryption/Decryption (address column) ---
        print("\n--- ChaCha20-Poly1305 Encryption/Decryption of 'address' column ---")
        # Generate key on driver and broadcast it
        chacha_key = spark.sparkContext.broadcast(driver_cipher_suite.generate_chacha20_key())

        # Define UDF for ChaCha20 encryption - CipherSuite instantiated within lambda
        chacha_encrypt_udf = udf(lambda addr: CipherSuite().encrypt_chacha20(addr, chacha_key.value),
                                 StructType([
                                     StructField("nonce", BinaryType(), True),
                                     StructField("ciphertext", BinaryType(), True),
                                     StructField("tag", BinaryType(), True)
                                 ]))

        df_chacha_encrypted = df.withColumn("address_chacha_encrypted", chacha_encrypt_udf(df["address"]))
        print("DataFrame with ChaCha20 Encrypted Address:")
        df_chacha_encrypted.select("id", "address_chacha_encrypted").show(truncate=False)

        # Define UDF for ChaCha20 decryption - CipherSuite instantiated within lambda
        # Access elements by index [0] for nonce, [1] for ciphertext, [2] for tag
        chacha_decrypt_udf = udf(lambda encrypted_tuple: \
                                  CipherSuite().decrypt_chacha20(
                                      encrypted_tuple[0],  # nonce
                                      encrypted_tuple[1],  # ciphertext
                                      encrypted_tuple[2],  # tag
                                      chacha_key.value
                                  ), StringType())

        df_chacha_decrypted = df_chacha_encrypted.withColumn("address_chacha_decrypted", chacha_decrypt_udf(df_chacha_encrypted["address_chacha_encrypted"]))
        print("DataFrame with ChaCha20 Decrypted Address (and original for comparison):")
        df_chacha_decrypted.select("id", "address", "address_chacha_decrypted").show(truncate=False)

        # Verify decryption
        df_chacha_decrypted.createOrReplaceTempView("chacha_data")
        match_count_chacha = spark.sql("SELECT COUNT(*) FROM chacha_data WHERE address = address_chacha_decrypted").collect()[0][0]
        print(f"ChaCha20 Decryption Verification: {match_count_chacha}/{total_count} rows match original. {'SUCCESS' if match_count_chacha == total_count else 'FAILURE'}\n")

        # --- RSA-OAEP Encryption/Decryption (address column) ---
        print("\n--- RSA-OAEP Encryption/Decryption of 'address' column ---")
        # For RSA, keys are large and should generally be managed externally.
        # Generate key pair on driver, get PEM bytes
        rsa_private_key_pem, rsa_public_key_pem = driver_cipher_suite.generate_rsa_key_pair(key_size=2048)

        # Broadcast public key PEM bytes for encryption UDF
        rsa_public_key_broadcast = spark.sparkContext.broadcast(rsa_public_key_pem)
        # Private key PEM bytes will be passed directly into the decryption UDF lambda
        # (This is less ideal for security in large clusters, but works around pickling issues)

        # Define UDF for RSA encryption - CipherSuite instantiated within lambda
        # It will load the public key from PEM bytes within the UDF
        rsa_encrypt_udf = udf(lambda addr: CipherSuite().encrypt_rsa(addr, rsa_public_key_broadcast.value), BinaryType())

        # Note: RSA has a plaintext size limit. If your address strings are too long,
        # this will raise a ValueError from `encrypt_rsa`.
        try:
            df_rsa_encrypted = df.withColumn("address_rsa_encrypted", rsa_encrypt_udf(df["address"]))
            print("DataFrame with RSA Encrypted Address:")
            df_rsa_encrypted.select("id", "address_rsa_encrypted").show(truncate=False)

            # Define UDF for RSA decryption - CipherSuite instantiated within lambda.
            # It will load the private key from PEM bytes within the UDF.
            rsa_decrypt_udf = udf(lambda encrypted_bytes: CipherSuite().decrypt_rsa(encrypted_bytes, rsa_private_key_pem), StringType())

            df_rsa_decrypted = df_rsa_encrypted.withColumn("address_rsa_decrypted", rsa_decrypt_udf(df_rsa_encrypted["address_rsa_encrypted"]))
            print("DataFrame with RSA Decrypted Address (and original for comparison):")
            df_rsa_decrypted.select("id", "address", "address_rsa_decrypted").show(truncate=False)

            # Verify decryption
            df_rsa_decrypted.createOrReplaceTempView("rsa_data")
            match_count_rsa = spark.sql("SELECT COUNT(*) FROM rsa_data WHERE address = address_rsa_decrypted").collect()[0][0]
            print(f"RSA Decryption Verification: {match_count_rsa}/{total_count} rows match original. {'SUCCESS' if match_count_rsa == total_count else 'FAILURE'}\n")

        except ValueError as e:
            print(f"RSA encryption/decryption demo skipped due to plaintext length constraint: {e}")
            print("Please try with shorter 'address' strings for the RSA example.")
        except Exception as e:
            print(f"An unexpected error occurred during RSA demo: {e}")

        spark.stop()
        print("Spark Session Stopped.")



from google.oauth2 import service_account
from google.auth.transport.requests import Request
from googleapiclient.discovery import build
from googleapiclient.http import MediaIoBaseDownload
import pandas as pd
import base64
import pickle
from typing import Optional, Dict, Any, Union
import json
import os
import io

class XGoogleDriveReader:
    """Class to handle Google Drive operations and read Google Sheets"""
    
    SCOPES = ['https://www.googleapis.com/auth/drive',
              'https://www.googleapis.com/auth/spreadsheets']
    
    def __init__(self, auth_file_path: str = 'auth/service-account-gdrive.json',
                 scopes: list = SCOPES):
        """
        Initialize GoogleDriveReader with service account credentials
        
        Args:
            auth_file_path: Path to service account JSON file (relative or absolute)
        """
        try:
            # Get absolute path if relative path provided
            if not os.path.isabs(auth_file_path):
                current_dir = os.path.dirname(os.path.abspath(__file__))
                auth_file_path = os.path.join(current_dir, auth_file_path)
            
            # Load credentials from JSON file
            with open(auth_file_path, 'r') as f:
                self.credentials = json.load(f)
                
            self.creds = None
            self.service = None
            self.sheets_service = None
            self.scopes = scopes
            
        except Exception as e:
            raise ValueError(f"Failed to load authentication file: {str(e)}")

    def authenticate(self) -> bool:
        """Authenticate with Google Drive API using service account"""
        try:

            self.creds = service_account.Credentials.from_service_account_info(
                self.credentials,
                scopes=self.scopes
            )
            
            self.service = build('drive', 'v3', credentials=self.creds)
            self.sheets_service = build('sheets', 'v4', credentials=self.creds)
            print("Successfully authenticated with service account")
            return True
            
        except Exception as e:
            print(f"Authentication failed: {str(e)}")
            return False
    
    def get_file_id(self, file_path: str, mime_type: str = None) -> Optional[str]:
        """
        Get Google Drive file ID by path
        
        Args:
            file_path: Full path to the file in Google Drive (e.g. 'folder1/folder2/filename')
            mime_type: Optional MIME type filter
            
        Returns:
            str: File ID if found, None otherwise
        """
        try:
            path_parts = file_path.strip('/').split('/')
            file_name = path_parts[-1]
            parent_path = '/'.join(path_parts[:-1])
            
            # Build query
            query = f"name='{file_name}'"
            if mime_type:
                query += f" and mimeType='{mime_type}'"
            
            # Add parent folder constraints if path provided
            if parent_path:
                parent_id = self._get_folder_id(parent_path)
                if parent_id:
                    query += f" and '{parent_id}' in parents"
                    
            results = self.service.files().list(
                q=query,
                fields="files(id, name)").execute()
            items = results.get('files', [])
            
            if not items:
                print(f"File '{file_path}' not found")
                return None
                
            return items[0]['id']
            
        except Exception as e:
            print(f"Error getting file ID: {str(e)}")
            return None

    def _get_folder_id(self, folder_path: str) -> Optional[str]:
        """
        Get folder ID from path
        
        Args:
            folder_path: Path to folder (e.g. 'folder1/folder2')
            
        Returns:
            str: Folder ID if found, None otherwise
        """
        try:
            current_parent = 'root'
            
            for folder in folder_path.split('/'):
                query = f"name='{folder}' and mimeType='application/vnd.google-apps.folder'"
                if current_parent != 'root':
                    query += f" and '{current_parent}' in parents"
                    
                results = self.service.files().list(
                    q=query,
                    fields="files(id)").execute()
                items = results.get('files', [])
                
                if not items:
                    print(f"Folder '{folder}' not found in path '{folder_path}'")
                    return None
                    
                current_parent = items[0]['id']
                
            return current_parent
            
        except Exception as e:
            print(f"Error getting folder ID: {str(e)}")
            return None

    def read_sheet(self, 
                   file_path: str, 
                   sheet_name: Optional[str] = None,
                   range_name: Optional[str] = None) -> Optional[pd.DataFrame]:
        """
        Read Google Sheet into pandas DataFrame
        
        Args:
            file_path: Full path to the file in Google Drive (e.g. 'folder1/folder2/filename')
            sheet_name: Optional sheet name (default: first sheet)
            range_name: Optional range to read (default: all)
            
        Returns:
            pd.DataFrame: DataFrame containing sheet data
        """
        try:
            # Get file ID - Updated method name
            file_id = self.get_file_id(file_path)
            if not file_id:
                return None
            
            # Get sheet properties
            sheet_metadata = self.sheets_service.spreadsheets().get(
                spreadsheetId=file_id).execute()
            
            if not sheet_name:
                sheet_name = sheet_metadata['sheets'][0]['properties']['title']
            
            # Construct range
            range_str = f"'{sheet_name}'"
            if range_name:
                range_str += f"!{range_name}"
            
            # Read data
            result = self.sheets_service.spreadsheets().values().get(
                spreadsheetId=file_id,
                range=range_str).execute()
            
            values = result.get('values', [])
            if not values:
                print("No data found in sheet")
                return None
            
            # Convert to DataFrame
            df = pd.DataFrame(values[1:], columns=values[0])
            return df
            
        except Exception as e:
            print(f"Error reading sheet: {str(e)}")
            return None

    def download_file(self, 
                     file_path: str, 
                     local_path: str,
                     mime_type: str = None) -> bool:
        """
        Download file from Google Drive
        
        Args:
            file_path: Full path to file in Google Drive (e.g. 'folder1/folder2/filename')
            local_path: Local path where to save the file
            mime_type: Optional MIME type for export (for Google Docs, Sheets etc.)
            
        Returns:
            bool: True if download successful, False otherwise
        """
        try:
            # Get file ID using the new method
            file_id = self.get_file_id(file_path)
            if not file_id:
                print(f"File not found: {file_path}")
                return False

            # Create local directory if it doesn't exist
            os.makedirs(os.path.dirname(local_path), exist_ok=True)

            # Download the file
            if mime_type:
                # Export Google Workspace files
                request = self.service.files().export_media(
                    fileId=file_id,
                    mimeType=mime_type
                )
            else:
                # Download regular files
                request = self.service.files().get_media(
                    fileId=file_id
                )

            # Stream the file content
            fh = io.BytesIO()
            downloader = MediaIoBaseDownload(fh, request)
            done = False

            while not done:
                status, done = downloader.next_chunk()
                if status:
                    print(f"Download progress: {int(status.progress() * 100)}%")

            # Save the file
            fh.seek(0)
            with open(local_path, 'wb') as f:
                f.write(fh.read())
                f.flush()

            print(f"File downloaded successfully to: {local_path}")
            return True

        except Exception as e:
            print(f"Error downloading file: {str(e)}")
            return False

    def list_files(self, folder_path: str = '') -> Optional[list]:
        """
        List all files and subfolders in the specified Google Drive folder
        
        Args:
            folder_path: Path to folder in Google Drive (e.g. 'folder1/folder2')
                        Empty string '' means root folder
            
        Returns:
            list: List of dictionaries containing file/folder information with structure:
                  [{'name': str, 'id': str, 'type': str, 'path': str}, ...]
            None: If error occurs
        """
        try:
            # Get folder ID for the path
            folder_id = 'root' if not folder_path else self._get_folder_id(folder_path)
            if folder_path and not folder_id:
                return None

            # Build query to list all files/folders in the directory
            query = f"'{folder_id}' in parents and trashed=false"
            
            # Get list of all files/folders
            results = []
            page_token = None
            
            while True:
                response = self.service.files().list(
                    q=query,
                    spaces='drive',
                    fields='nextPageToken, files(id, name, mimeType)',
                    pageToken=page_token
                ).execute()
                
                for file in response.get('files', []):
                    file_type = 'folder' if file['mimeType'] == 'application/vnd.google-apps.folder' else 'file'
                    file_path = os.path.join(folder_path, file['name']) if folder_path else file['name']
                    
                    results.append({
                        'name': file['name'],
                        'id': file['id'],
                        'type': file_type,
                        'path': file_path
                    })
                    
                    # Recursively list files in subfolders
                    if file_type == 'folder':
                        sub_files = self.list_files(file_path)
                        if sub_files:
                            results.extend(sub_files)
                
                page_token = response.get('nextPageToken')
                if not page_token:
                    break
            
            return results
            
        except Exception as e:
            print(f"Error listing files: {str(e)}")
            return None

# Example usage
def read_sheet_from_gdrive():
    # Initialize reader with default auth file path
    reader = XGoogleDriveReader()
    
    # Or specify a different path
    # reader = XGoogleDriveReader('path/to/different/auth.json')
    
    # Authenticate
    if reader.authenticate():
        df = reader.read_sheet(
            file_path="DSWD-Sheets/Chill",
            sheet_name="0825_DS",
            range_name="B2:E16"
        )
        
        if df is not None:
            print(df.head())

def download_file_from_gdrive():
    reader = XGoogleDriveReader("/home/hadoop/tests/hpmeis-7784af30c873.json")
    
    if reader.authenticate():
        # Download Google Sheet as Excel
        success = reader.download_file(
            file_path="HPMEIS/AICS.xlsx",
            local_path="/home/hadoop/poc/my_sheet.xlsx"
        )
    print(f"Download success: {success}")

def list_files_from_gdrive():
    """Example usage of list_files function"""
    reader = XGoogleDriveReader("/home/hadoop/tests/hpmeis-7784af30c873.json")
    
    if reader.authenticate():
        # List all files in a specific folder
        files = reader.list_files("HPMEIS")
        
        if files:
            print("\nFiles found:")
            for file in files:
                prefix = "📁" if file['type'] == 'folder' else "📄"
                print(f"{prefix} {file['path']}")
                
if __name__ == "__main__":
    # read_sheet_from_gdrive()
    list_files_from_gdrive()
    download_file_from_gdrive()


from .main import XGoogleDriveReader as gdr
from ...tools.logger import Logger as logger
from pyspark.sql import SparkSession
from pyspark.sql.types import StructType

class XGoogleService:
    def __init__(self, dataStorage: dict = None, readOptions: dict = None, schema: StructType = None,
                 sparkSession: SparkSession = None):
        self.dataStorage = dataStorage
        self.readOptions = readOptions
        self.schema = schema
        # Initialize Spark session
        self.spark = sparkSession

    def processRequest(self):
        """
        Process the request to get data from Google Sheets.
        
        Returns:
            pyspark.sql.DataFrame: Spark DataFrame if available, otherwise None.
        """
        dataStorage = self.dataStorage
        readOptions = self.readOptions
        file_path = dataStorage.get("file_path")
        sheet_name = dataStorage.get("sheet_name", None)
        range_name = dataStorage.get("range_name", None)
        auth_file_path = dataStorage.get("auth_file_path", None)

        # Initialize Google Drive Reader
        if auth_file_path:
            reader = gdr(auth_file_path)
        else:
            reader = gdr()  # Use default auth file path

        # Authenticate
        if reader.authenticate():
            if dataStorage.get("format").upper().endswith("SHEETS"):
                print(f'GoogleService[processRequest]: Google Sheets executing sheet reading.')
                pandas_df = reader.read_sheet(
                    file_path=file_path,
                    sheet_name=sheet_name,
                    range_name=range_name
                )
                if pandas_df is not None:
                    # Convert pandas DataFrame to Spark DataFrame
                    try:
                        spark_df = self.spark.createDataFrame(
                            pandas_df,
                            schema=self.schema
                        )
                        logger().log(f'GoogleService[processRequest]: Spark DataFrame created with schema: {spark_df.schema}', level='debug')
                        return spark_df
                    except Exception as e:
                        logger().log(f'GoogleService[processRequest]: Failed to convert to Spark DataFrame: {str(e)}', level='error')
                        return None
                else:
                    logger().log('GoogleService[processRequest]: No data found in the specified sheet.', level='warning')
                    return None
            elif dataStorage.get("format").upper().endswith("FILE"):
                # Implement Google Drive File Download format handling if needed
                print(f'GoogleService[processRequest]: Google Drive File Download executing download.')
                print(f'GoogleService[processRequest]: file_path={file_path}')
                print(f'GoogleService[processRequest]: local_path={readOptions.get("local_path", "downloaded_file")}')
                reader.download_file(
                    file_path=file_path,
                    local_path=readOptions.get("local_path", "downloaded_file"),
                    mime_type=readOptions.get("mime_type", None)
                )
            else:
                logger().log('GoogleService[processRequest]: Unsupported format specified.', level='error')
                return None
        else:
            logger().log('GoogleService[processRequest]: Authentication failed.', level='error')
            return None



from langchain.agents import AgentExecutor, create_tool_calling_agent
from langchain.prompts import ChatPromptTemplate
from langchain_openai import ChatOpenAI
from llama_index.embeddings.openai import OpenAIEmbedding
from llama_index.core import VectorStoreIndex, SimpleDirectoryReader, StorageContext, load_index_from_storage, Settings
from llama_index.core.tools import QueryEngineTool, ToolMetadata
from xursparks.error.main import GenericXursparksException

class NLPAgent:

    def __init__(self, temperature:float = None,
                llm_ver:str = None,
                model:str = None,
                embed_batch_size:int = None,
                streaming:bool = None,
                verbose:bool = None) -> None:
        self.temperature = temperature
        self.llm_ver = llm_ver
        self.model = model
        self.embed_batch_size = embed_batch_size
        self.streaming = streaming
        self.verbose = verbose

    def get_agent(self,
                  callback_handler = None,
                  persist_path:str = None,
                  file_path:str = None,
                  metadata_name:str = None,
                  metadata_desc:list[str] = None,
                  system_context:str = None,
                  return_intermediate_steps:bool = None,
                  handle_parsing_errors:bool = None,
                  max_iterations:int = None):
       
        if metadata_name is None or metadata_name.strip() == '':
            metadata_name = "xurpas_faq"
        if metadata_desc is None or metadata_desc.strip() == '':
            metadata_desc = ["Provides information about Xurpas."
                        "Use a detailed plain text question as input to the tool."]
        if system_context is None or system_context.strip() == '':
            raise GenericXursparksException("Must Provide System Context")
        if return_intermediate_steps is None:
            return_intermediate_steps = True
        if handle_parsing_errors is None:
            handle_parsing_errors = True
        if max_iterations is None:
            max_iterations = 10

        if callback_handler != None:
            print("With callback handler!")
            llm = ChatOpenAI(
                temperature=self.temperature,
                model_name=self.llm_ver,
                streaming=self.streaming,
                callbacks=callback_handler
            )
        else:
            llm = ChatOpenAI(
                temperature=self.temperature,
                model=self.llm_ver,
                verbose=self.verbose,
                streaming=self.streaming
            )

        embed_model = OpenAIEmbedding(
            model=self.model, embed_batch_size=self.embed_batch_size
        )

        Settings.llm = llm
        Settings.embed_model = embed_model

        try:
            storage_context = StorageContext.from_defaults(
                persist_dir=persist_path
            )
            data_index = load_index_from_storage(storage_context)

            index_loaded = True
            print("Index was already created. We just loaded it from the local storage.")
        except:
            index_loaded = False
            print("Index is not present. We need it to create it again.")

        if not index_loaded:
            print("Creating Index..")

            docs = SimpleDirectoryReader(
                file_path
            ).load_data()
            data_index = VectorStoreIndex.from_documents(docs)
            data_index.storage_context.persist(persist_dir=persist_path)

            index_loaded = True

        query_engine = data_index.as_query_engine(similarity_top_k=3)

        query_engine_tools = [
            QueryEngineTool(
                query_engine=query_engine,
                metadata=ToolMetadata(
                    name=metadata_name,
                    description=(
                        metadata_desc
                    ),
                ),
            )
        ]

        converted_tools = [t.to_langchain_tool() for t in query_engine_tools]

        print("No of LlamaIndex Tools:", len(converted_tools))

        prompt = ChatPromptTemplate.from_messages(
            [
                (
                    "system",
                    system_context,
                ),
                ("placeholder", "{chat_history}"),
                ("human", "{input}"),
                ("placeholder", "{agent_scratchpad}"),
            ]
        )

        agent = create_tool_calling_agent(
            llm=llm,
            tools=converted_tools,
            prompt=prompt
        )

        agent_executor = AgentExecutor(
            agent=agent,
            tools=converted_tools,
            verbose=self.verbose,
            return_intermediate_steps=True,
            handle_parsing_errors=True,
            max_iterations=10
        )

        return agent_executor
    



from .main import Xurl
from ...tools.logger import Logger

class APIService:
    def __init__(self, base_url: str = None,
                 args: dict = None):
        self.base_url = base_url
        self.args = args

    def processRequest(self):
        """
        Get the response from the args dictionary.
        Returns:
            str: The response if available, otherwise None.
        """
        base_url = self.base_url
        xurl_json = self.args.get("json")
        xurl_auth = self.args.get("auth")
        if xurl_auth is None:
            xurl_auth = self.args.get("ntlm_auth")
        xurl_headers = self.args.get("headers")
        xurl_method = self.args.get("method", "GET")
        xurl_verify_ssl = self.args.get("verify_ssl", False)
        # print("************** [xursparks.services.http.api_service.processRequest] START **************")
        # print(f'base_url: {base_url}')
        # print(f'xurl_json: {xurl_json}')
        # print(f'xurl_auth: {xurl_auth}')
        # print(f'xurl_method: {xurl_method}')
        # print(f'xurl_verify_ssl: {xurl_verify_ssl}')
        # print(f'xurl_headers: {xurl_headers}')
        # print("************** [xursparks.services.http.api_service.processRequest] END **************")
        xurl = Xurl(
            url=base_url,
            headers=xurl_headers,
            json=xurl_json,
            auth=xurl_auth,
            verify_ssl=xurl_verify_ssl,
        )
        print(f'xurl[processRequest]: {str(xurl)}')
        if xurl_method.upper() == "POST":
            response = xurl.post()
        else:
            response = xurl.get()
        Logger().log(f'xurl[processRequest]: {response}', level='debug')
        return response




class GenericXursparksException(Exception):
    """Custom exception for Xursparks Generic Errors."""
    pass

class ApiServiceException(Exception):
    """Custom exception for API service errors."""
    
    def __init__(self, message, status_code=None):
        super().__init__(message)
        self.status_code = status_code

    def __str__(self):
        return f"ApiServiceException: {self.args[0]} (Status Code: {self.status_code})" if self.status_code else f"ApiServiceException: {self.args[0]}"


import configparser
import smtplib, ssl
import subprocess
import requests
import json
import re
import pandas as pd
import datetime
import time
import boto3
import os
import mimetypes
from requests_ntlm import HttpNtlmAuth
from os.path import abspath
from pyspark.sql import SparkSession
from pyspark.sql import HiveContext
from pyspark.sql.types import *
from pyspark.sql import DataFrame as SparkDataFrame
from email.utils import formataddr
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from dateutil.relativedelta import relativedelta
from datetime import date as dt, timedelta
from urllib.parse import quote
from email.mime.base import MIMEBase
from email import encoders
from xurpas_data_quality import DataReport
from .services.itemreader import ItemReader
from .tools.excelreader import FileReader
from .services.http.main import Xurl
from .services.http.api_reader import APIReader
from .tools.encryption import CipherSuite
from .services.google import XGoogleService as xgservice

debug_mode: bool
job_type: str

def initialize(args):
	#initialize global variables
	global jobArguments
	global customJobArgs
	global targetTableDataset
	global writeDate
	global processDate
	global targetTable
	global targetTableOptions
	global sourceTableOptions
	global jdbcOptions
	global sourceTables
	global sparkConfig
	global sparkDefaultOptions
	global auditOptions
	global startTimestamp
	global optionalParams
	global startTimestamp
	global startTimeEpoch
	global scope_env
	global debug_mode
	global job_type

	jobArguments = {}
	customJobArgs = {}
	targetTableOptions = {}
	sourceTableOptions = {}
	jdbcOptions = {}
	sourceTables = {}
	sparkConfig = {}
	sparkDefaultOptions = {}
	auditOptions = {}
	optionalParams = {}
	startTimestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	startTimeEpoch = time.perf_counter()
	scope_env = "DEV"
	debug_mode = False
	job_type = "default"

	print("--------------------")
	print("-      Job Args    -")
	print("--------------------")
	for i in range(len(args)):
		jobArg = args[i]
		if(i == 0):
			jobArguments["py-file"] = jobArg
		else:
			print(str(i) + " : " + str(jobArg))
			item = jobArg.replace("--", "")
			key = item.split("=")[0]
			value = item.split("=")[1]
			jobArguments[str(key)] = value
			if(item.startswith("customJobArgs")):
				customArg = item.replace("customJobArgs=", "").split("=",2)
				customJobArgs[customArg[0]] = customArg[1]
	
	# checking for new arg from custom
	if "debug" in jobArguments is not None and jobArguments["debug"]:
		debug_mode = True
	if "job_type" in jobArguments:
		job_type = jobArguments["job_type"].strip()

	# setup appName / processDate / writeDate
	global appName
	appName = jobArguments["client-id"] + "." + jobArguments["target-table"]
	processDate = datetime.datetime.strptime(jobArguments["process-date"], "%Y-%m-%d").date()
	writeDate = datetime.datetime.strptime(jobArguments["process-date"], "%Y-%m-%d").date()
	
	#setup properties reader
	global jobProperties
	jobProperties = configparser.ConfigParser()
	jobProperties.read(jobArguments["properties-file"])
	
	#setup global configurations
	x = callAPI(apiURL=getDefaultJobProperties("global.config"), httpMode="get").json()
	jobGlobalConfigurations = x["message"]
	for item in jobGlobalConfigurations["global_parameters"]:
		key = item["key"]
		if(key.startswith("spark.") or key.startswith("session.")):
			if(key.startswith("session.")):
				key = key.replace("session.", "", 1)
			sparkConfig[key] = item["value"]
			if(item["key"].startswith("spark")):
				sparkDefaultOptions[item["key"]] = item["value"]
		elif(item.get("key").startswith("audit")):
			auditOptions[item.get("key").replace("audit.", "")] = item.get("value")
		if(key.startswith("environment")):
			scope_env = item["value"].upper()
	
	#setup target table dictionary
	contextApi = getDefaultJobProperties("job.context") + jobArguments.get("client-id") + "/" + jobArguments.get("target-table").replace(".", "/")
	x = callAPI(apiURL=contextApi, httpMode="get").json()
	jobTables = x["message"]
	targetTable = {}
	
	if("storage_type" in jobTables.keys()):
		storageType = jobTables["storage_type"]
		if("AWS" in storageType.upper()):
			targetTable["storage_type"] = "s3a://"
		elif("GCS" in storageType.upper()):
			targetTable["storage_type"] = "gs://"
		elif("HDFS" in storageType.upper()):
			targetTable["storage_type"] = "hdfs://"
	else:
		targetTable["storage_type"] = "s3a://"

	targetTable["table_name"] = jobTables["table_name"]
	targetTable["database_name"] = jobTables["database_name"]
	if "resource" in jobTables.keys():
		targetTable["resource"] = jobTables["resource"]
	else:
		targetTable["resource"] = ""
	targetTable["folder"] = jobTables["folder"] + "/"
	if "local" in jobArguments["master"]:
		targetTable["path"] = jobTables["folder"]
	else:
		targetTable["path"] = targetTable["storage_type"] + targetTable["resource"] + jobTables["folder"]
	targetTable["write_mode"] = jobTables["write_mode"]
	targetTable["redistribution"] = jobTables["redistribution"]
	targetTable["partition_name"] = jobTables["partition_name"]
	targetTable["partition_date_format"] = jobTables["partition_date_format"]
	targetTable["format"] = jobTables["format"]
	targetTable["number_of_partitions"] = jobTables["number_of_partitions"]

	if("optional_parameters" in jobTables.keys()):
		for param in jobTables["optional_parameters"]:
			key = param.get("key")
			optionalParams[key] = param.get("value")
			if(key == "options.target"):
				targetTableOptions = json.loads(param.get("value"))
			elif(key.startswith("options.target.")):
				key = key.replace("options.target.","")
				targetTableOptions[key] = param.get("value")
			elif(key.startswith("options.jdbc.")):
				key = key.replace("options.jdbc.","")
				jdbcOptions[key] = param.get("value")
			elif(key.startswith("options.source.")):
				key = key.replace("options.source.","")
				sourceTableOptions[key] = json.loads(param.get("value"))
		jdbcOptions["dbtable"] = targetTable["table_name"]

	if("dependencies" in jobTables.keys()):
		for item in jobTables["dependencies"]:
			sjc = item["source_job_context"]
			source = {}
			if("storage_type" in sjc.keys()):
				storageType = sjc["storage_type"]
				if("AWS" in storageType.upper()):
					source["storage_type"] = "s3a://"
				elif("GCS" in storageType.upper()):
					source["storage_type"] = "gs://"
				elif("HDFS" in storageType.upper()):
					source["storage_type"] = "hdfs://"
			else:
				source["storage_type"] = "s3a://"

			tableName = sjc["table_name"]
			source["table_name"] = tableName
			source["database_name"] = sjc["database_name"]
			if "resource" in sjc.keys():
				source["resource"] = sjc["resource"]
			else:
				source["resource"] = ""
			source["path"] = source["storage_type"] + source["resource"] + sjc["folder"]
			source["write_mode"] = sjc["write_mode"]
			source["redistribution"] = sjc["redistribution"]
			source["partition_name"] = sjc["partition_name"]
			source["partition_date_format"] = sjc["partition_date_format"]
			source["format"] = sjc["format"]
			source["number_of_partitions"] = sjc["number_of_partitions"]
			source["read_start_date"] = parseDatePattern(item["read_start_date"], processDate)
			source["read_end_date"] = parseDatePattern(item["read_end_date"], processDate)
			source["read_mode"] = item["read_mode"]
			source["search_tolerance"] = int(item["search_tolerance"])
			sourceTables[tableName] = source

	if("keytab" in jobArguments.keys() and "principal" in jobArguments.keys()):
		print(f'creating session with keytab : {jobArguments.get("keytab")} and principal: {jobArguments.get("principal")}')
		sparkConfig["spark.kerberos.keytab"] = jobArguments["keytab"]
		sparkConfig["spark.kerberos.principal"] = jobArguments["principal"]

	initializeSparkSession()

def initializeSparkSession():
	print("SETTING UP SparkSession")
	global sparkSession
	try:
		print(f'sparkConfig: {sparkConfig}')
		if "local" not in jobArguments["master"]:
			sparkSession = SparkSession.builder \
								.appName(appName) \
								.master(jobArguments["master"]) \
								.config(map=sparkConfig) \
								.enableHiveSupport() \
								.getOrCreate()
		else:
			sparkSession = SparkSession.builder \
								.appName(appName) \
								.getOrCreate()
	except Exception as e:
		raise e

def readSourceTableParquet(s3_path):
	print("PASOK SA BANGA!")
	try:
		df = sparkSession.read.parquet(s3_path)
		return df
	except Exception as e:
		print(e)

def loadSourceTable(**kwargs):
	dataStorage = None
	schema = None
	readOptions = {}
	processDate = None
	columnFilter = None
	if("dataStorage" in kwargs.keys()):
		dataStorage = kwargs.get("dataStorage")
	if("schema" in kwargs.keys()):
		schema = kwargs.get("schema")
	if("readOptions" in kwargs.keys()):
		readOptions = kwargs.get("readOptions")
	if("processDate" in kwargs.keys()):
		processDate = kwargs.get("processDate")
	if("columnFilter" in kwargs.keys()):
		columnFilter = kwargs.get("columnFilter")
		
	return readSourceTable(dataStorage, schema, readOptions, processDate, columnFilter)

def readSourceTable(dataStorage, schema, readOptions, processDate, columnFilter):
	sparkSession = getSparkSession()
	if(schema == None):
		schema = StructType([])
	else:
		readOptions['customSchema'] = schema
	columns = schema
	print(f'dataStorage[235]: {str(dataStorage)}')
	sourceTableOption = sourceTableOptions.get(dataStorage.get("table_name"))
	if(sourceTableOption == None):
		sourceTableOption = {}
	df = sparkSession.createDataFrame([], schema=columns)

	if "local" not in jobArguments["master"]:
		options = {**sparkDefaultOptions, **sourceTableOption, **readOptions}
	else:
		options = {**sourceTableOption, **readOptions}

	if (dataStorage.get("format").upper().startswith("GOOGLE")):
		# Implement reading from Google Sheets
		df = xgservice(dataStorage=dataStorage, readOptions=readOptions, schema=schema, sparkSession=sparkSession).processRequest()
	elif (dataStorage.get("format").upper() == "API"):
		mode = readOptions.get("mode", "default")
		if(mode == "ntlm"):
			source_datastorage = {}
			source_datastorage["base_url"] = readOptions.get("source_base_url")
			source_datastorage["headers"] = readOptions.get("source_headers")
			source_datastorage["method"] = readOptions.get("source_method", "GET")

			auth_user = readOptions.get("source_auth_user")
			auth_password = readOptions.get("source_auth_password")
			source_datastorage["auth"] = HttpNtlmAuth(auth_user, auth_password)
			
			response_key = readOptions.get("source_response_key", "value")
			
			#print all source_datastorage keys and values
			print(f"Source Storage: {source_datastorage}")

			ar_response = APIReader(args=source_datastorage).fetch_data()
			
			return ar_response
		else:
			token_datastorage = {}
			token_datastorage["base_url"] = readOptions.get("token_base_url")
			token_datastorage["auth"] = readOptions.get("token_auth")
			token_datastorage["headers"] = readOptions.get("token_headers")
			token_datastorage["method"] = readOptions.get("token_method", "GET")
			token_key_name = readOptions.get("token_key_name", "token")

			#print all token_datastorage keys and values
			print(f"Token Data Storage: {token_datastorage}")

			ar_response = APIReader(args=token_datastorage).fetch_data()	
			
			#print ar_response
			print(f"ar_response[token_datastorage]: {str(ar_response)}")

			source_datastorage = {}
			source_datastorage["base_url"] = readOptions.get("source_base_url")
			source_datastorage["auth"] = readOptions.get("source_auth")
			source_datastorage["method"] = readOptions.get("source_method", "GET")
			source_datastorage["headers"] = readOptions.get("source_headers")
			source_datastorage["key"] = f"Bearer {ar_response.get(token_key_name)}"
			response_key = readOptions.get("source_response_key", "value")

			#print all source_datastorage keys and values
			print(f"Source Storage: {source_datastorage}")

			ar_response = APIReader(args=source_datastorage).fetch_data()
			#print ar_response
			# print(f"ar_response[source_datastorage]: {str(ar_response)}")	

			df = APIReader(spark_session=sparkSession, data=ar_response.get(response_key)).process_data(df=df)
		
	else:	# for non-api formats
		if("basePath" not in options.keys()):
			options["basePath"] = dataStorage["path"]
		print("Reading : " + options["basePath"])

		if("read_start_date" not in dataStorage.keys() or "read_end_date" not in dataStorage.keys() or processDate == None):
			if(len(schema) == 0):
				df = sparkSession.read.options(**options).format(dataStorage["format"]).load()
			else:
				df = sparkSession.read.format(dataStorage["format"]).options(**options).schema(schema).load()
		elif(dataStorage.get("format").upper() == "JDBC"):
			print("READING SOURCE as JDBC")
			if(len(schema) > 0):
				print("READING WITH SCHEMA")
				df = sparkSession.read.schema(schema).format(dataStorage["format"]).options(**options).load()
			else:
				print("READING WITHOUT SCHEMA")
				df = sparkSession.read.format(dataStorage["format"]).options(**options).load()
		else:
			if(dataStorage["read_mode"] == "STRICT"):
				df = readTableInStrictMode(dataStorage, schema, options, df)
			if(dataStorage["read_mode"] == "SEARCH"):
				df = readTableInSearchMode(dataStorage, schema, options, df)
			if(dataStorage["read_mode"] == "LAX"):
				df = readTableInLaxMode(dataStorage, schema, options, df)

		if(columnFilter != None):
			df = df.filter(columnFilter)

		options.pop("basePath")

	return df

def generateTablePaths(dataStorage, options):
	generatedTablePaths = []

	if(dataStorage["read_start_date"] != None and dataStorage["read_end_date"] != None and dataStorage["read_start_date"] != dataStorage["read_end_date"]):
		print("GENERATING PATHS FOR TABLE: " + dataStorage["table_name"])
	else:
		print("GENERATING PATH FOR TABLE: " + dataStorage["table_name"])

	startDate = dataStorage.get("read_start_date").strftime("%Y-%m-%d")

	if(dataStorage.get("read_mode") == "SEARCH"):
		endDate = (dataStorage.get("read_start_date") - timedelta(days=int(dataStorage.get("search_tolerance")))).strftime("%Y-%m-%d")
	else:
		endDate = dataStorage.get("read_end_date").strftime("%Y-%m-%d")

	sdate = dt(int(startDate.split("-")[0]), \
				int(startDate.split("-")[1]), \
				int(startDate.split("-")[2]))   # start date
	edate = dt(int(endDate.split("-")[0]), \
				int(endDate.split("-")[1]), \
				int(endDate.split("-")[2]))   # end date

	dateRange = pd.date_range(edate, sdate)
	for date in pd.date_range(edate, sdate):
		generatedTablePaths.append(produceTablePath(dataStorage) + date.strftime(dataStorage["partition_date_format"]) + "/")

	return reversed(generatedTablePaths)

def readTableInStrictMode(dataStorage, schema, options, dataFrame):
	print("Reading Table in STRICT mode.")
	sparkSession = getSparkSession()
	paths = generateTablePaths(dataStorage, options)
	for path in paths:
		if "excel" in dataStorage.get('format'):
			path = path + options.get('filename')
		print("READING Partition(Strict): " + path)
		try:
			if(len(schema) == 0):
				df = sparkSession.read.options(**options) \
						.format(dataStorage["format"]).load(path)
			else:
				df = sparkSession.read.options(**options) \
						.format(dataStorage["format"]).schema(schema).load(path)
		except Exception as e:
			raise Exception("Partition Path:" + path + " does not exist")
		if(dataFrame.rdd.isEmpty()):
			dataFrame = df
		else:
			dataFrame = dataFrame.union(df)
	return dataFrame

def readTableInSearchMode(dataStorage, schema, options, dataFrame):
	print("Reading Table in SEARCH mode.")
	sparkSession = getSparkSession()
	sdate = dataStorage["read_start_date"] - \
			pd.DateOffset(days=dataStorage["search_tolerance"])
	print(f"format: {dataStorage.get('format')}")
	for path in generateTablePaths(dataStorage, options):
		if "excel" in dataStorage.get('format'):
			path = path + options.get('filename')
		print("READING Partition(Search): " + path)
		try:
			if(len(schema) == 0):
				dataFrame = sparkSession.read\
						.format(dataStorage["format"])\
						.options(**options).load(path)
			else:
				dataFrame = sparkSession.read.options(**options) \
						.format(dataStorage["format"]).schema(schema).load(path)
			if(dataFrame != None):
				print("Path: " + path + " exists.")
				break
		except Exception as e:
			print("Path: " + path + " does not exist.")
	return dataFrame

def readTableInLaxMode(dataStorage, schema, options, dataFrame):
	print("Reading Table in LAX mode.")
	sparkSession = getSparkSession()
	if(dataStorage.get("read_start_date") == "*" and dataStorage.get("read_end_date") == "*"):
		if(len(schema) == 0):
			dataFrame = sparkSession.read.options(**options) \
				.format(dataStorage["format"]).load(dataStorage["path"])
		else:
			dataFrame = sparkSession.read.options(**options) \
				.format(dataStorage["format"]).schema(schema).load(dataStorage["path"])
	else:
		paths = generateTablePaths(dataStorage, options)
		for path in paths:
			if "excel" in dataStorage.get('format'):
				path = path + options.get('filename')
			print("READING Partition(LAX): " + path)
			try:
				if(len(schema) == 0):
					df = sparkSession.read.options(**options) \
							.format(dataStorage["format"]).load(path)
				else:
					df = sparkSession.read.options(**options) \
							.format(dataStorage["format"]).schema(schema).load(path)
			except Exception as e:
				print("Partition Path: " + path + " does not exist")
			if(dataFrame.rdd.isEmpty()):
				dataFrame = df
			else:
				dataFrame = dataFrame.union(df)
	return dataFrame

def writeToTargetTable(targetTable, writeDate):
	print("WRITING TO TargetTable")
	path = getDefaultJobProperties("target.table.path") + getWriteOptionProperties("partition") + "=" + writeDate + "/"
	if(getWriteOptionProperties("redistribution").upper() == "REPARTITION"):
		targetTable = targetTable.repartition(int(getWriteOptionProperties("num.partition")))
	if(getWriteOptionProperties("redistribution").upper() == "COALESCE"):
		targetTable = targetTable.coalesce(int(getWriteOptionProperties("num.partition")))
	else:
		targetTable = targetTable.repartition(int(getWriteOptionProperties("num.partition")))

	try:
		targetTable.write.format(getWriteOptionProperties("format")) \
			.mode(getWriteOptionProperties("mode")) \
			.save(path)
	except Exception as e:
		raise e
		
	print("Done writing to target table")

def writeToTargetTable(dataStorage, dataset, writeOptions, jdbcWriteOptions, processDate):
	if "local" not in getJobArguments(key="master"):
		options = {**sparkDefaultOptions, **writeOptions}
	else:
		options = {**writeOptions}

	path = ""
	if(dataStorage.get("format").upper() == "JDBC"):
		jdbcOptions["dbtable"] = dataStorage.get("table_name")
		options = {**jdbcOptions, **jdbcWriteOptions, **options}
		print("WRITING as JDBC")
		dataset.write.options(**options).format(dataStorage.get("format"))\
			.mode(dataStorage.get("write_mode")).save()
	else:
		path = generateTablePath(dataStorage)

		if(dataStorage.get("redistribution").upper() == "REPARTITION"):
			dataset = dataset.repartition(int(dataStorage.get("number_of_partitions")))
		else:
			dataset = dataset.coalesce(int(dataStorage.get("number_of_partitions")))

		lower_case_dict = {k.lower(): v for k, v in options.items()}
		if("filename" in lower_case_dict):
			path = path + options.get("filename")
		print("WRITING to path: " + path)
		print(f"options: {str(options)}")
		dataset.write.options(**options).format(dataStorage.get("format"))\
			.mode(dataStorage.get("write_mode")).save(path)
	
	print(f"targetTableDataStorage: {str(dataStorage)}")

def writeData(**kwargs):
	#TODO - implement targettable options
	global writeDate
	global targetTable
	global targetTableOptions
	global targetTableDataset
	global jdbcOptions
	global startTimeEpoch
	global startTimestamp
	global endTimestamp
	global endTimeEpoch
	if("targetTable" in kwargs.keys()):
		thisTargetTable = kwargs.get("targetTable")
		targetTable = thisTargetTable	
	if("writeDate" in kwargs.keys()):
		thisWriteDate = kwargs.get("writeDate")
		writeDate = thisWriteDate	
	if("targetTableOptions" in kwargs.keys()):
		thisTargetTableOptions = kwargs.get("targetTableOptions")
		targetTableOptions = {**targetTableOptions, **thisTargetTableOptions}
	if("targetTableDataset" in kwargs.keys()):
		thisTargetTableDataset = kwargs.get("targetTableDataset")
		targetTableDataset = thisTargetTableDataset
	if("jdbcOptions" in kwargs.keys()):
		thisJdbcOptions = kwargs.get("jdbcOptions")
		jdbcOptions = {**jdbcOptions, **thisJdbcOptions}
	writeToTargetTable(targetTable, targetTableDataset, targetTableOptions, jdbcOptions, writeDate)
	endTimestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	endTimeEpoch = time.perf_counter()
	elapsed_time = endTimeEpoch - startTimeEpoch
	print(f"start : {startTimestamp} | {startTimeEpoch}")
	print(f"end : {endTimestamp} | {endTimeEpoch}")
	print(f"elapsed : {elapsed_time}")

def doJobAuditing(data):
	smtp_server = auditOptions["server"]
	port =  auditOptions["port"]
	sender_email =  auditOptions["sender"]
	password =  auditOptions["password"]
	maskname =  auditOptions["maskname"]
	subject =  "Audit Report: " + appName.upper()
	recipients =  data["recipients"]
	email_recipients = recipients.split(",")
	email_content =  data["content"]

	# Create a secure SSL context
	msg = MIMEMultipart("alternative")
	msg['Subject'] = subject
	msg['From'] = formataddr((maskname, sender_email))
	msg['To'] = ", ".join(email_recipients)
	html_part = MIMEText(email_content, "html")
	msg.attach(html_part)

	context = ssl.create_default_context()
	with smtplib.SMTP_SSL(smtp_server, port, context=context) as server:
		server.login(sender_email, password)
		server.sendmail(sender_email, email_recipients, msg.as_string())

def getDefaultJobProperties(propertyName):
	return getSpecificJobProperties("default", propertyName)

def getWriteOptionProperties(propertyName):
	return getSpecificJobProperties("write_options", propertyName)

def getJdbcOptionProperties(propertyName):
	return getSpecificJobProperties("jdbc_options", propertyName)

def getJobAuditProperties(propertyName):
	return getSpecificJobProperties("job_audit", propertyName)

def getSpecificJobProperties(groupName, propertyName):
	return jobProperties.get(groupName, propertyName)

def getDictionarFromProperties(groupName):
	return dict(jobProperties.items(groupName))

def getSparkSession():
	return sparkSession

def setTargetTableDataset(targetTable):
	global targetTableDataset
	if isinstance(targetTable, pd.DataFrame):
		targetTableDataset = sparkSession.createDataFrame(targetTable)
	elif isinstance(targetTable, SparkDataFrame):
		targetTableDataset = targetTable
	else:
		raise TypeError("Unknown DataFrame type.")

def setProcessDate(process_date):
	global processDate
	processDate = process_date

def tearDown():
	print("stopping sparkSession")
	sparkSession.stop

def executeSubProcess(options):
	print("------------ExecuteSubProcess---------------")
	print("options: " + str(options))
	return subprocess.run([options["command"]], \
				shell=options["shell"], \
				capture_output=options["capture_output"], \
				text=options["text"]).stdout

def handlePostWrite():
	print("------------HandlePostWrite---------------")
	global endTimestamp
	command = "ps aux | grep -w " + auditOptions.get("user") + "| " \
				+ "grep -w " + appName + "| " \
				+ "grep -v grep | cut -d' ' -f4"
	print("command: " + command)
	auditOptions["command"] = command
	auditOptions["shell"] = True
	auditOptions["capture_output"] = True
	auditOptions["text"] = True
	usages = executeSubProcess(auditOptions).split()
	usage = "Memory Usage: " + usages[0] + "<br />" + "CPU Usage :" + usages[1]
	count = targetTableDataset.count()
	content = auditOptions["content"].replace("#COUNT#", str(count)).replace("#PROCESS_DATE#", processDate.strftime("%Y-%m-%d")) \
				.replace("#USAGE#", usage)
	auditOptions["content"] = content
	doJobAuditing(auditOptions)
	endTimestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
	usageAPI = str(getDefaultJobProperties("usage.logs")) \
				+ "?name=Spark-Dev" \
				+ "&job_name=" + quote(appName) \
				+ "&cpu_usage=" + quote(str(usages[1])) \
				+ "&memory_usage=" + quote(str(usages[0])) \
				+ "&message=" + quote("Audit Report") \
				+ "&start=" + quote(str(startTimestamp)) \
				+ "&end=" + quote(str(endTimestamp))
	callAPI(apiURL=usageAPI, httpMode="post")

def getJobArguments(**kwargs):
	key = kwargs.get("key")
	if(key != None and len(key) > 0):
		return jobArguments.get(key)
	else:
		return jobArguments

def getCustomJobArguments(**kwargs):
	key = kwargs.get("key")
	if(key != None and len(key) > 0):
		return customJobArgs.get(key)
	else:
		return customJobArgs

def getSourceTables(**kwargs):
	key = kwargs.get("key")
	if(key != None and len(key) > 0):
		return sourceTables.get(key)
	else:
		return sourceTables

def getTargetTable():
	return targetTable

def setTargetTable(dataStorage):
	global targetTable
	targetTable = dataStorage

def getProcessDate():
	return processDate

def getWriteDate():
	return writeDate

def generateTablePath(dataStorage):
    partition_names = dataStorage["partition_name"].lower().split(",")
    partition_format = dataStorage["partition_date_format"]
    # Use processDate or writeDate as needed; here, using writeDate for consistency
    partition_value = writeDate.strftime(partition_format)

    # Build partition path dynamically
    partition_path = ""
    for pname in partition_names:
        partition_path += f"{pname}={partition_value}/"

    path = f'{dataStorage["path"]}{partition_path}'
    return path

def produceTablePath(dataStorage):
	path = dataStorage["path"] + dataStorage["partition_name"].lower() + "="
	return path

def parseBaseDate(datePattern, processDate):
	if(datePattern == "*"):
		return None
	else:
		if("CURRENT_DATE" not in datePattern and "TRANSACTION_DATE" not in datePattern):
			raise Exception("date pattern must start with CURRENT_DATE or TRANSACTION_DATE")
		elif("CURRENT_DATE" in datePattern and "TRANSACTION_DATE" in datePattern):
			raise Exception("date pattern must start with CURRENT_DATE or TRANSACTION_DATE")
		else:
			if("TRANSACTION_DATE" in datePattern):
				return processDate
			return datetime.datetime.now()

def parseDatePattern(datePattern, processDate):
	derivedDate = datetime.datetime.now()

	if("TRANSACTION" in datePattern or "CURRENT" in datePattern):
		derivedDate = parseBaseDate(datePattern, processDate)
		if(derivedDate == None):
			return None
	else:
		derivedDate = datetime.datetime.now().strftime(datePattern)

	pattern = re.findall("PLUS_YEARS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						+ pd.DateOffset(years=offset)

	pattern = re.findall("MINUS_YEARS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						- pd.DateOffset(years=offset)

	pattern = re.findall("PLUS_MONTHS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						+ pd.DateOffset(months=offset)

	pattern = re.findall("MINUS_MONTHS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						- pd.DateOffset(months=offset)

	pattern = re.findall("PLUS_WEEKS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						+ pd.DateOffset(weeks=offset)

	pattern = re.findall("MINUS_WEEKS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						- pd.DateOffset(weeks=offset)

	if("FIRST" in datePattern):
		derivedDate = derivedDate.replace(day=1)
	if("LAST" in datePattern):
		derivedDate = derivedDate + relativedelta(day=31)

	pattern = re.findall("PLUS_DAYS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						+ pd.DateOffset(days=offset)

	pattern = re.findall("MINUS_DAYS_\\d+", datePattern)
	if(len(pattern) > 0):
		offset = int(pattern[0].split("_")[2])
		derivedDate = pd.to_datetime(datetime.datetime.now()) \
						- pd.DateOffset(days=offset)

	return derivedDate

def getSparkDefaultOptions():
	return sparkDefaultOptions

def setWriteDate(thisDate:datetime):
	global writeDate
	writeDate = thisDate

def callAPI(**kwargs):
	apiURL = kwargs.get("apiURL", None)
	httpMode = kwargs.get("httpMode", None)
	body = kwargs.get("body", None)
	headerToken = getDefaultJobProperties("api.token")
	if(httpMode.upper() == "GET"):
		return requests.get(apiURL, headers={'Authorization': headerToken})
	elif(httpMode.upper() == "POST"):
		if(body != None and len(body) > 0):
			return requests.post(apiURL, headers={'Authorization': headerToken}, json=body)
		else:
			return requests.post(apiURL, headers={'Authorization': headerToken})
	else:
		raise Exception("Invalid httpMode.")

def getSourceTableOption(tableName):
	return sourceTableOptions.get(tableName)

def getTargetTableOption():
	return targetTableOptions

def setTargetTableOption(options):
	targetTableOptions = {**options, **targetTableOptions}

def getOptionalParams(**kwargs):
	key = kwargs.get("key", None)
	if(key != None and len(key) > 0):
		return optionalParams.get(key)
	else:
		return optionalParams

def getSparkConfig(**kwargs):
	key = kwargs.get("key", None)
	if(key != None and len(key) > 0):
		return sparkConfig.get(key)
	else:
		return sparkConfig
	
def uploadToS3(**kwargs):
	# Replace these placeholder values with your own
	filePath = kwargs.get("filePath")
	filePaths = kwargs.get("filePaths")
	bucketName = kwargs.get("bucketName")
	subDirectory = kwargs.get("subDirectory")
	extraArgs = kwargs.get("contentType")

	if(extraArgs == None or extraArgs.strip() == ''):
		extraArgs = "application/octet-stream"
	if(subDirectory == None):
		subDirectory = targetTable.get("folder") \
			+ targetTable.get("partition_name").lower() + "=" \
			+ writeDate.strftime(targetTable.get("partition_date_format")) \
			+ "/"
			
	if(bucketName == None):
		bucketName = targetTable.get("resource")
	if(filePaths == None):
		filePaths = []
		filePaths.append(filePath)

	sparkConfig = getSparkConfig()
	ACCESS_KEY = sparkConfig.get("session.spark.hadoop.fs.s3a.access.key")
	SECRET_KEY = sparkConfig.get("session.spark.hadoop.fs.s3a.secret.key")
	partitionName = getTargetTable().get("partition_name")

	# Create an S3 client
	s3 = boto3.client('s3', aws_access_key_id=ACCESS_KEY, aws_secret_access_key=SECRET_KEY)

	# Loop through file names and upload each file to S3
	for filename in filePaths:
		# Get the file path of the file
		abspath = os.path.abspath(filename)
		# Get the base name of the file
		theFile = os.path.basename(filename)

		# Get the content type of the file
		content_type = mimetypes.guess_type(filename)[0]
		resource = f"{subDirectory}{theFile}"

		# Upload the file to S3 in the specified subdirectory
		with open(filename, 'rb') as f:
			s3.upload_fileobj(f, bucketName, resource, ExtraArgs={'ContentType': extraArgs})
	
def sendEmail(**kwargs):
	print("kwargs: " + str(kwargs))
	smtp = auditOptions["server"]
	smtp_port =  auditOptions["port"]
	from_email =  auditOptions["sender"]
	email_password =  auditOptions["password"]

	subject = str(kwargs.get("subject"))
	to_email = kwargs.get("to_email")
	message = str(kwargs.get("message"))
	files = kwargs.get("files")
	rename = kwargs.get("rename")
	bodyContentType = kwargs.get("body_content_type")
	msg = MIMEMultipart()
	msg['Subject'] = subject
	msg['From'] = from_email
	msg['To'] = ', '.join(to_email)

	# Attach message
	if(bodyContentType != None and len(bodyContentType) > 0):
		body = MIMEText(message, bodyContentType)
	else:
		body = MIMEText(message)

	msg.attach(body)

	#check if email has attachment
	if(files != None and len(files) > 0):
		# Attach Documents
		if(rename == None):
			for file in files:
				with open(file, 'rb') as attachment_file:
					part = MIMEBase('application', "octet-stream")
					part.set_payload(open(file, "rb").read())
				encoders.encode_base64(part)
				part.add_header('Content-Disposition', 'attachment', filename=str(file))  
				msg.attach(part)
		else:
			for file, new_name in zip(files, rename):
				with open(file, 'rb') as attachment_file:
					part = MIMEBase('application', "octet-stream")
					part.set_payload(open(file, "rb").read())
				encoders.encode_base64(part)
				part.add_header('Content-Disposition', f'attachment; filename= {new_name}')  
				msg.attach(part)

	# Send email
	s = smtplib.SMTP(smtp, smtp_port)
	s.ehlo()
	s.starttls()
	s.login(from_email, email_password)
	s.sendmail(from_email, to_email, msg.as_string())
	s.quit()

def fetchFileFromS3():
	# Set the AWS access key and secret key
	ACCESS_KEY = sparkConfig.get("session.spark.hadoop.fs.s3a.access.key")
	SECRET_KEY = sparkConfig.get("session.spark.hadoop.fs.s3a.secret.key")

	# Set the name of the S3 bucket, the subdirectory, and the file to be loaded
	bucket_name = targetTable.get("resource")
	subdirectory_name = targetTable.get("folder")[1:-1] \
		+ targetTable.get("partition_name").lower() + "=" \
		+ writeDate.strftime(targetTable.get("partition_date_format"))

	# Get the directory of the current file
	current_dir = os.path.dirname(os.path.realpath(__file__))
	
	# Create the path to the subdirectory
	subdir_path = os.path.join(current_dir, "private")
	
	# Create the subdirectory if it doesn't exist
	if not os.path.exists(subdir_path):
		os.makedirs(subdir_path)

	# Create an S3 resource
	s3 = boto3.resource('s3', aws_access_key_id=ACCESS_KEY, aws_secret_access_key=SECRET_KEY)
	bucket = s3.Bucket(bucket_name)
	for obj in bucket.objects.filter(Prefix=subdirectory_name):
		file_key = obj.key
		file_name = file_key.split('/')[-1]
		index_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "private")
		local_file_path = os.path.join(index_path, file_name)
		bucket.download_file(file_key, local_file_path)

	return f'{local_file_path}'

def tapChatGPT(**kwargs):
	OPENAI_API_KEY = kwargs.get("key")
	DATA_STATISTICS = kwargs.get("statistics")
	QUERY = kwargs.get("query")
	MODEL = kwargs.get("model")
	MAX_TOKENS = kwargs.get("maxTokens")

	if(MODEL == None):
		MODEL = "gpt-3.5-turbo-instruct"

	if(MAX_TOKENS == None):
		MAX_TOKENS = 100

	# Define the API endpoint
	url = f'https://api.openai.com/v1/chat/completions'

	messages = [
		{"role": "system", "content": "You are a helpful assistant."},
		{"role": "user", "content": f'{QUERY}: {DATA_STATISTICS}'},
	]

	# Define the headers for the API request
	headers = {
				"Content-Type": "application/json",
				"Authorization": "Bearer " + OPENAI_API_KEY
			}

	# Define the data for the API request
	data = {
			"model": MODEL,
			"messages": messages,
			"max_tokens": MAX_TOKENS
		}

	# Make the API request
	response = requests.post(url, headers=headers, data=json.dumps(data))

	# Get the response text
	response_text = response.json()

	return response_text

def updateS3Permission(**kwargs):
	permission = kwargs.get("permission")
	subdirectory_name = kwargs.get("subDirectory")

	if(permission == None):
		permission = 'public-read'
	if(subdirectory_name == None):
		subdirectory_name = targetTable.get("folder") \
			+ targetTable.get("partition_name").lower() + "=" \
			+ writeDate.strftime(targetTable.get("partition_date_format"))

	# Set the AWS access key and secret key
	ACCESS_KEY = sparkConfig.get("session.spark.hadoop.fs.s3a.access.key")
	SECRET_KEY = sparkConfig.get("session.spark.hadoop.fs.s3a.secret.key")
	s3 = boto3.client('s3', aws_access_key_id=ACCESS_KEY, aws_secret_access_key=SECRET_KEY)

	# Set the name of the S3 bucket, the subdirectory, and the file to be loaded
	bucket_name = targetTable.get("resource")

	# Use paginator to handle buckets with many objects
	paginator = s3.get_paginator('list_objects_v2')
	pages = paginator.paginate(Bucket=bucket_name, Prefix=subdirectory_name)

	# Iterate over each page and each object in the page
	for page in pages:
		if 'Contents' in page:
			for obj in page['Contents']:
				s3.put_object_acl(ACL=permission, Bucket=bucket_name, Key=obj['Key'])
				print(f"Permissions updated for {obj['Key']}")

def generateDataQualityReport(**kwargs):
	title = kwargs.get("dq_title")
	filePath = kwargs.get("dq_file_path")
	df = kwargs.get("dq_df")
	message = kwargs.get("dq_message")

	if(title == None):
		title = f'{targetTable.get("database_name")}_{targetTable.get("table_name")}'
	if(filePath == None):
		filePath = f'{os.path.dirname(__file__)}/{targetTable.get("database_name")}_{targetTable.get("table_name")}.html'
	if(df == None):
		df = targetTableDataset.toPandas()
	if(message == None):
		message = 'Hi'
	
	# integration with XAIL DQ reporting library
	report = DataReport(df=df,
		report_name=title, 
		file_path=f'{filePath}')
	report.to_file()

	# read the HTML file
	with open(filePath, 'r') as f:
		html_content = f.read()
	recipients = auditOptions.get("recipients")
	if(optionalParams.get("audit.recipients") != None and len(optionalParams.get("audit.recipients")) > 0):
		recipients = f'{recipients},optionalParams.get("audit.recipients")'
	emailRecipients = recipients.split(",")
	files = [filePath]
	rename = [f'{title}.html']
	sendEmail(subject=f'[{scope_env}]Data Profile[{title}]-{processDate}', message=message, from_email=auditOptions.get("sender"), 
		to_email=emailRecipients, files=files, rename=rename)

def createHiveTable():
	hive_database = targetTable.get("database_name")
	hive_table = targetTable.get("table_name")
	partition_column = targetTable.get("partition_name").lower()
	partition_value = processDate.strftime(targetTable.get("partition_date_format"))
	hdfs_path = generateWritePath()

	# Generate column types based on DataFrame schema
	column_types = ", ".join([f"{col} {get_hive_data_type(data_type)}" for col, data_type in targetTableDataset.dtypes])

	hiveContext = HiveContext(sparkSession)
	hive_query = f"""
				CREATE TABLE IF NOT EXISTS `{hive_table}` (
					{column_types}
				) 
				PARTITIONED BY ({partition_column} string)
				STORED AS PARQUET
				LOCATION '{targetTable["path"]}
				USING hive'
			"""
	print(f'hive query(create if not exist): {hive_query}')
	hiveContext.sql(hive_query)

def registerPartition(**kwargs):
	print(f'******************** Registering Partition to Hive *******************')
	mode = kwargs.get("mode")
	hive_database = targetTable.get("database_name")
	hive_table = targetTable.get("table_name")
	partition_name = kwargs.get("partition_name")
	partition_names = kwargs.get("partition_names")  # Expecting list of tuples: [("col1", "val1"), ("col2", "val2")]
	partition_location = kwargs.get("partition_location", generateWritePath())

	if mode is None:
		mode = ""

	if partition_name is None and partition_names is None:
		partition_name = targetTable.get("partition_name").lower()
		partition_spec = f"{partition_name}='{writeDate.strftime(targetTable.get('partition_date_format'))}'"
	else:
		if partition_name or partition_names is None:
			partition_spec = f"{partition_name}='{writeDate.strftime(targetTable.get('partition_date_format'))}'"
		else:
			# Build partition spec string for multiple columns using the tuple list
			if partition_names:
				partition_spec = ",".join([f"{col}='{val}'" for col, val in partition_names])
			else:
				# fallback to old logic if not provided
				partition_cols = targetTable.get("partition_name").lower().split(",")
				partition_format = targetTable.get("partition_date_format")
				partition_value = processDate.strftime(partition_format)
				partition_spec = ",".join([f"{col}='{partition_value}'" for col in partition_cols])

	hive_query = f"""
		ALTER TABLE {hive_database}.{hive_table}
		ADD IF NOT EXISTS PARTITION ({partition_spec})
		LOCATION '{partition_location}'
	"""
	print(f'hive query(add partition): {hive_query}')
	sparkSession.sql(hive_query)
	# hiveContext = HiveContext(sparkSession)
	# hiveContext.sql(hive_query)

def generateWritePath():
	partition_names = targetTable["partition_name"].lower().split(",")
	partition_format = targetTable["partition_date_format"]
	partition_value = writeDate.strftime(partition_format)

	# Build partition path dynamically
	partition_path = ""
	for pname in partition_names:
		partition_path += f"{pname}={partition_value}/"
	
	path = f'{targetTable["path"]}{partition_path}'
	return path

def generateWritePaths():
    """
    Generate partition paths dynamically:
    - First partition uses writeDate
    - Succeeding partitions use unique values from targetTableDataset columns
    Returns a list of partition paths.
    """
    partition_names = targetTable["partition_name"].lower().split(",")
    partition_format = targetTable["partition_date_format"]
    base_path = targetTable["path"]
    paths = []

    if len(partition_names) == 1:
        # Only one partition column, use writeDate
        partition_value = writeDate.strftime(partition_format)
        partition_path = f"{partition_names[0]}={partition_value}/"
        paths.append(f"{base_path}{partition_path}")
    else:
        # First partition uses writeDate, others use DataFrame columns
        first_partition = partition_names[0]
        other_partitions = partition_names[1:]
        first_value = writeDate.strftime(partition_format)

        # Get unique combinations of the other partition columns
        unique_combinations = (
            targetTableDataset
            .select(*other_partitions)
            .dropDuplicates()
            .toPandas()
            .to_dict(orient="records")
        )

        for combo in unique_combinations:
            partition_path = f"{first_partition}={first_value}/"
            for col in other_partitions:
                partition_path += f"{col}={combo[col]}/"
            paths.append(f"{base_path}{partition_path}")

    return paths

# Function to map PySpark data types to Hive data types
def get_hive_data_type(spark_data_type):
	result = ""
	if isinstance(spark_data_type, StringType):
		result = "STRING"
	elif isinstance(spark_data_type, IntegerType):
		result = "INT"
	elif isinstance(spark_data_type, DoubleType):
		result = "DOUBLE"
	elif isinstance(spark_data_type, TimestampType):
		result = "TIMESTAMP"
	else:
		# Add more mappings for other data types as needed
		result = "STRING"
	return result



from langchain_openai import ChatOpenAI
from pydantic import BaseModel, Field
from langchain.schema import SystemMessage
from langchain.agents import initialize_agent, AgentType
from langchain.tools import BaseTool
from datetime import date
from typing import Type
import json
import pandas as pd
from typing import Optional
from sqlalchemy import create_engine

class HRAgent:

    def __init__(self, temperature:float = None,
                llm_ver:str = None,
                model:str = None,
                embed_batch_size:int = None,
                streaming:bool = None,
                verbose:bool = None,
                path:str = None,
                db_domain:str = None,
                db_port:str = None,
                db_name:str = None,
                db_password:str = None) -> None:
        self.temperature = temperature
        self.llm_ver = llm_ver
        self.model = model
        self.embed_batch_size = embed_batch_size
        self.streaming = streaming
        self.verbose = verbose
        self.path = path
        self.db_domain = db_domain
        self.db_port = db_port
        self.db_name = db_name
        self.db_password = db_password

    def get_agent(self,last_message=None, callback_handler=None,
                  name:str = None,
                  employee_name:str = None,
                  companies:list[str] = None,
                  content:str = None):
        name = name
        employee_name = employee_name
        year = date.today().year
        print(f'GET AGENT YEAR={year}')
        companies = companies
        print(f'COMPANIES={companies}')
        content=content
        # if (employee_name != None) and (len(employee_name) > 0):
        #     content +=f"\n5/ You are currently talking about the employee {employee_name}"
        if last_message:
            content +=f"\n5/ The last response you sent was: \"{last_message[1]}\""
        print(content)
        system_message = SystemMessage(content)
        hr_tool = HRRecordsTool(
            path = self.path,
            db_domain=self.db_domain,
            db_port=self.db_port,
            db_name=self.db_name,
            db_password=self.db_password
        )
        leaves_tool = LeavesTool(
            path = self.path,
            db_domain=self.db_domain,
            db_port=self.db_port,
            db_name=self.db_name,
            db_password=self.db_password
        )
        employee_list_tool = EmployeeListTool(
            path = self.path,
            db_domain=self.db_domain,
            db_port=self.db_port,
            db_name=self.db_name,
            db_password=self.db_password
        )

        tools = [
            hr_tool,
            leaves_tool,
            employee_list_tool
        ]

        agent_kwargs = {
            "system_message": system_message,
        }
        llm_ver = self.llm_ver
        if callback_handler != None:
            llm=ChatOpenAI(temperature=self.temperature, model_name=llm_ver, streaming=self.streaming, callbacks=callback_handler)
        else:
            llm = ChatOpenAI(temperature=0, model=llm_ver, verbose=self.verbose)
        agent = initialize_agent(
            tools,
            llm,
            agent=AgentType.OPENAI_FUNCTIONS,
            verbose=self.verbose,
            agent_kwargs=agent_kwargs,
        )

        return agent

class HRRecordsInput(BaseModel):
    """Inputs for get_content function"""
    first_name: str = Field(description="The first name of the employee to get records for.")
    last_name: str = Field(description="The last name of the employee to get records for.")
    start_date: date = Field(description="The start date of the entries that you want to query. Set to January 1, 1970 if not specifically asked for a joining date.")
    end_date: date = Field(description="The end date of the entries that you want to query. Set to the end of the current year if not specifically asked for a joining date.")
    

class HRRecordsTool(BaseTool):
    name = "hr_records_tool"
    description = "Use to get information about an employee. Whenever you see a record_url in your result, render it as a link."
    args_schema: Type[BaseModel] = HRRecordsInput
    
    # Make the database fields optional with a default of None
    path: Optional[str] = Field(None, description="The config path")
    db_domain: Optional[str] = Field(None, description="The database domain")
    db_port: Optional[str] = Field("3306", description="The database port")
    db_name: Optional[str] = Field(None, description="The database name")
    db_password: Optional[str] = Field(None, description="The database password")
    #doc: Optional[str] = Field(None, description="The employee doc")
    #emp: Optional[str] = Field(None, description="The emp unique id")

    def get_data(self, first_name=None, last_name=None, start_date=None, end_date=None):

        if start_date == 'None':
            start_date = None
        if end_date == 'None':
            end_date = None
        if len(first_name) > 0:
            first_name = first_name.split()[0]
        if len(last_name) > 0:
            last_name = last_name.split()[-1]
            print(f'{first_name} {last_name} {start_date} {end_date}')
            conn_string = self.get_conn_string()
            db_connection = create_engine(conn_string)
        
            query = """
            SELECT * FROM `tabEmployee`
            WHERE first_name LIKE %s
            AND last_name LIKE %s
            """
            params = (f"{first_name}%", f"{last_name}%")
            try:
                df = pd.read_sql(query, con=db_connection, params=params)

                if not df.empty:
                    df['full_name'] = df['first_name'] +' '+df['last_name']
                    df['record_url'] = df.apply(lambda x: f"/app/employee/{x['name']}", axis=1)
                    return df.to_dict(orient='records')
                else:
                    return "Employee Not Found"
            except Exception as e:
                print(f"Error executing query: {e}")
                return str(e)
        
    def get_conn_string(self):
        path = f'{self.path}/site_config.json'
        with open(path) as file:
            j = json.load(file)
            #print(j)
            conn_string = f'mariadb+pymysql://{j["db_name"]}:{j["db_password"]}@{self.db_domain}:{self.db_port}/{j["db_name"]}'
            #print("conn_string: ", conn_string)
            return conn_string

    def _run(self, first_name: str, last_name: str, start_date: date, end_date : date):
        #return self.get_data(first_name, last_name, start_date, end_date)
        data = self.get_data(first_name, last_name, start_date, end_date)

        if data != "Employee Not Found":
            if data != None:
                print("Employee Data:", data)
                emp = data[0]
                full_name = emp.get('full_name', None)
                print("Employee Full Name:", full_name)
                return {"employee": data, "full_name": full_name}
            else:
                return "No data Found. "
        else:
            #print(data)
            return {"error": data}

    def _arun(self,  first_name: str, last_name: str, attr: str):
        raise NotImplementedError("error here")
    

class LeavesInput(BaseModel):
    """Inputs for get_content function"""
    first_name: str = Field(description="The first name of the employee to get records for.")
    last_name: str = Field(description="The last name of the employee to get records for.")
    start_date: date = Field(description="The start date of the entries that you want to query")
    end_date: date = Field(description="The end date of the entries that you want to query")


class LeavesTool(BaseTool):
    name = "leaves_tool"
    description = "Use to get information about an employee's leaves. Use only if specifically asked about leaves."
    args_schema: Type[BaseModel] = LeavesInput

    # Make the database fields optional with a default of None
    path: Optional[str] = Field(None, description="The config path")
    db_domain: Optional[str] = Field(None, description="The database domain")
    db_port: Optional[str] = Field("3306", description="The database port")
    db_name: Optional[str] = Field(None, description="The database name")
    db_password: Optional[str] = Field(None, description="The database password")


    def get_data(self, first_name, last_name, start_date=None, end_date=None):
        print(f'{first_name} {last_name} {start_date} {end_date}')
        #doc = self.doc
        #print(doc)
        # return "doc"
        conn_string = self.get_conn_string()
        db_connection = create_engine(conn_string)
    
        query = """
        SELECT name FROM `tabEmployee`
        WHERE first_name LIKE %s
        AND last_name LIKE %s
        """
        params = (f"{first_name}%", f"{last_name}%")
        try:
            df = pd.read_sql(query, con=db_connection, params=params)

            if not df.empty:
                name = df['name'][0] 
                print("name: ", name)
                if name:
                    leave_query = """
                    SELECT leave_type, from_date, to_date, status, total_leave_days
                      FROM `tabLeave Application`
                    WHERE employee = %s
                    """
                    leave_params = (name,) 
                    
                    try:
                        leave_df = pd.read_sql(leave_query, con=db_connection, params=leave_params)

                        if not leave_df.empty:

                            return leave_df.to_dict(orient='records')
                        else:
                            return "No Leave details found."
                    except Exception as e:
                        print(f"Error executing query: {e}")
                        return str(e)


                else:
                    return "Employee Not Found"
                
            else:
                return "Employee Not Found"
        except Exception as e:
            print(f"Error executing query: {e}")
            return str(e)

    def get_conn_string(self):
        path = f'{self.path}/site_config.json'
        with open(path) as file:
            j = json.load(file)
            conn_string = f'mariadb+pymysql://{j["db_name"]}:{j["db_password"]}@{self.db_domain}:{self.db_port}/{j["db_name"]}'
            return conn_string

    def _run(self, first_name: str, last_name: str, start_date: date, end_date : date):
        return self.get_data(first_name, last_name, start_date, end_date)

    def _arun(self,  first_name: str, last_name: str, attr: str):
        raise NotImplementedError("error here")


class EmployeeListInput(BaseModel):
    """Inputs for get_content function"""
    company: str = Field(description="The company to get records for. Set to blank to a list for all companies.")
    start_date: date = Field(description="The start date of the entries that you want to query. Set to January 1, 1970 if not specifically asked for a joining date.")
    end_date: date = Field(description="The end date of the entries that you want to query. Set to the end of the current year if not specifically asked for a joining date.")
    status: str = Field(description="The employee status to look for. Possibe values are ['Active', 'Inactive', 'Suspended', 'Left']")
    

class EmployeeListTool(BaseTool):
    name = "employee_list_tool"
    description = "Use to get a list of employees for a company, Always provide full names. When presenting a list, include the record_url rendered as a link."
    args_schema: Type[BaseModel] = EmployeeListInput

    # Make the database fields optional with a default of None
    path: Optional[str] = Field(None, description="The config path")
    db_domain: Optional[str] = Field(None, description="The database domain")
    db_port: Optional[str] = Field("3306", description="The database port")
    db_name: Optional[str] = Field(None, description="The database name")
    db_password: Optional[str] = Field(None, description="The database password")

    def get_data(self, company, start_date=None, end_date=None, status='Active'):
        print(f'{company} {start_date} {end_date}')
        conn_string = self.get_conn_string()
        db_connection = create_engine(conn_string)
        conditions = f" WHERE status = '{status}'"
        if company:
            conditions = f" WHERE company = '{company}'"

        if start_date and start_date != 'None':
            if status == 'Left':
                conditions = f" WHERE relieving_date between '{start_date}' and '{end_date}'"
            else:
                conditions = f" WHERE date_of_joining between '{start_date}' and '{end_date}'"

        sql = f'SELECT name, first_name, last_name, company, status, designation, date_of_joining, relieving_date FROM `tabEmployee` {conditions} '
        print("sql: ", sql)
        df = pd.read_sql(sql, con=db_connection)

        if not df.empty:
            df['full_name'] = df['first_name'] +' '+df['last_name']
            df['record_url'] = df.apply(lambda x: f"/app/employee/{x['name']}", axis=1)
            return df.to_dict(orient='records')
        else:
            return 'No Records Found'

    def get_conn_string(self):
        path = f'{self.path}/site_config.json'
        with open(path) as file:
            j = json.load(file)
            conn_string = f'mariadb+pymysql://{j["db_name"]}:{j["db_password"]}@{self.db_domain}:{self.db_port}/{j["db_name"]}'
            return conn_string

    def _run(self, company: str, start_date: date, end_date : date, status: str):
        return self.get_data(company, start_date, end_date, status)

    def _arun(self,  company: str, attr: str):
        raise NotImplementedError("error here")


from typing import Dict, List, Union, Tuple
import pandas as pd
import openpyxl
from openpyxl.utils import get_column_letter
from dataclasses import dataclass

@dataclass
class HeaderConfig:
    """Configuration for Excel headers including merge information"""
    text: str
    start_col: Union[int, str]
    end_col: Union[int, str] = None
    start_row: int = 1
    end_row: int = None
    alignment: dict = None
    style: dict = None

class ExcelWriter:
    """Class to handle writing complex Excel files with merged headers"""
    
    def __init__(self, filename: str):
        self.filename = filename
        self.workbook = openpyxl.Workbook()
        self.current_sheet = None
        
    def create_sheet(self, name: str) -> None:
        """Create a new worksheet"""
        if name in self.workbook.sheetnames:
            self.workbook.remove(self.workbook[name])
        self.current_sheet = self.workbook.create_sheet(name)
        
    def write_complex_headers(self, 
                            headers: List[HeaderConfig],
                            start_row: int = 1) -> None:
        """
        Write complex headers with merging
        
        Args:
            headers: List of HeaderConfig objects defining the headers
            start_row: Starting row for headers
        """
        if not self.current_sheet:
            raise ValueError("No active sheet. Call create_sheet() first.")
            
        for header in headers:
            # Convert column letters to numbers if needed
            start_col = header.start_col if isinstance(header.start_col, int) else \
                       openpyxl.utils.column_index_from_string(header.start_col)
            
            end_col = header.end_col if isinstance(header.end_col, int) else \
                     openpyxl.utils.column_index_from_string(header.end_col) if header.end_col else start_col
                     
            end_row = header.end_row if header.end_row else header.start_row
            
            # Write header text
            self.current_sheet.cell(header.start_row, start_col, header.text)
            
            # Merge cells if needed
            if start_col != end_col or header.start_row != end_row:
                merge_range = f"{get_column_letter(start_col)}{header.start_row}:"
                merge_range += f"{get_column_letter(end_col)}{end_row}"
                self.current_sheet.merge_cells(merge_range)
            
            # Apply styles if provided
            cell = self.current_sheet.cell(header.start_row, start_col)
            if header.alignment:
                cell.alignment = openpyxl.styles.Alignment(**header.alignment)
            if header.style:
                for style_attr, style_value in header.style.items():
                    setattr(cell, style_attr, style_value)
                    
    def write_dataframe(self,
                       df: pd.DataFrame,
                       start_row: int,
                       start_col: Union[int, str] = 1,
                       index: bool = False) -> None:
        """
        Write pandas DataFrame to Excel below headers
        
        Args:
            df: Pandas DataFrame to write
            start_row: Starting row for DataFrame
            start_col: Starting column (number or letter)
            index: Whether to write index
        """
        if isinstance(start_col, str):
            start_col = openpyxl.utils.column_index_from_string(start_col)
            
        # Write DataFrame
        for i, col in enumerate(df.columns):
            self.current_sheet.cell(start_row, start_col + i, str(col))
            
        for i, row in enumerate(df.values):
            for j, val in enumerate(row):
                self.current_sheet.cell(start_row + i + 1, start_col + j, val)
                
    def save(self) -> None:
        """Save the Excel file"""
        if "Sheet" in self.workbook.sheetnames:
            self.workbook.remove(self.workbook["Sheet"])
        self.workbook.save(self.filename)

if __name__ == "__main__":
    # Sample DataFrame
    df = pd.DataFrame({
        'Col1': [1, 2, 3],
        'Col2': [4, 5, 6],
        'Col3': [7, 8, 9],
        'Col4': [10, 11, 12]
    })
    
    # Create writer
    writer = ExcelWriter('complex_output.xlsx')
    writer.create_sheet('Sheet1')
    
    # Define complex headers
    headers = [
        HeaderConfig(
            text="Main Header 1",
            start_col="A",
            end_col="B",
            start_row=1,
            end_row=1,
            alignment={'horizontal': 'center'},
            style={'font': openpyxl.styles.Font(bold=True)}
        ),
        HeaderConfig(
            text="Main Header 2",
            start_col="C",
            end_col="D",
            start_row=1,
            end_row=1,
            alignment={'horizontal': 'center'},
            style={'font': openpyxl.styles.Font(bold=True)}
        ),
        HeaderConfig(
            text="Sub Header 1",
            start_col="A",
            start_row=2,
            alignment={'horizontal': 'center'}
        ),
        HeaderConfig(
            text="Sub Header 2",
            start_col="B",
            start_row=2,
            alignment={'horizontal': 'center'}
        )
    ]
    
    # Write headers and data
    writer.write_complex_headers(headers)
    writer.write_dataframe(df, start_row=3)
    writer.save()


from xursparks.error.main import GenericXursparksException as gxe
import requests, configparser

class XursparksService:
    def __init__(self) -> None:
        pass

    def fetch(self,
              api_url:str = None,
              api_key:str = None,
              api_secret: str = None,
              type: str = None) -> dict:
        resp:dict

        if api_url is None:
            return gxe(f'api_url is None')

        headerToken = f'Token {api_key}:{api_secret}'
        resp = requests.get(api_url, headers={'Authorization': headerToken})

        if type is not None and type.strip().lower() == 'json':
            resp = resp.json()

        return resp
    
    def loadPropertiesFile(self,
                           properties_path:str) -> any:
        if properties_path is None:
            return gxe(f'properties_path is None')
        properties:any
        properties = configparser.ConfigParser()
        properties.read(properties_path)

        return properties
    
    def getPropertyFromFile(self,
                           properties_path:str,
                           property_key:str,
                           property_group:str = None) -> str:
        if properties_path is None:
            return gxe(f'properties_path is None')
        
        if property_key is None:
            return gxe(f'property_key is None')
        
        if property_group is None:
            property_group = "default"
        
        property_value:str
        properties = configparser.ConfigParser()
        properties.read(properties_path)
        property_value = properties.get(property_group, property_key)

        return property_value


from xursparks.services.http.api_service import APIService
import pandas as pd
from pyspark.sql.types import StructType, StructField, StringType, DoubleType, IntegerType, BooleanType

class APIReader:
    """
    A class to read and process API data.
    """

    def __init__(self, api_url: str = None,
                 args: dict = None,
                 spark_session: object = None,
                 data: object = None):
        self.api_url = api_url
        self.args = args
        self.spark_session = spark_session
        self.data = data

    def fetch_data(self):
        headers = self.args.get("headers")
        if headers and "Authorization" in headers:
            # Remove old Authorization if exists
            headers.pop("Authorization")
            
            # Get new Authorization value
            auth_value = self.args.get("key")
            if auth_value:
                # Add new Authorization with updated value
                headers["Authorization"] = f"{auth_value}"
                print(f"Updated Authorization header with new token")
        
        print(f"[xurspasrks.http.api_reader][fetch_data] Headers: {headers}")

        api_service = APIService(base_url=self.args.get("base_url"), args={
            "auth": self.args.get("auth"),
            "headers": headers,
            "method": self.args.get("method", "GET"),
            "verify_ssl": self.args.get("verify_ssl", False),
        })

        return api_service.processRequest()

    def process_data(self, df:object = None):
        """
        Process the fetched data and convert to Spark DataFrame.
        
        Args:
            df: Optional DataFrame to process
            
        Returns:
            pyspark.sql.DataFrame: Processed Spark DataFrame
        """
        if self.data:
            # Create Pandas DataFrame
            panda_frame = pd.DataFrame(self.data)
            
            # Create StructType schema
            schema_fields = []
            for column, dtype in panda_frame.dtypes.items():
                if dtype == 'object':
                    field_type = StringType()
                elif dtype == 'float64':
                    field_type = DoubleType()
                elif dtype == 'int64':
                    field_type = IntegerType()
                elif dtype == 'bool':
                    field_type = BooleanType()
                else:
                    field_type = StringType()  # Default to string
                    
                schema_fields.append(StructField(column, field_type, True))
            
            schema = StructType(schema_fields)
            
            # Create Spark DataFrame with schema
            try:
                spark_df = self.spark_session.createDataFrame(
                    panda_frame,
                    schema=schema
                )
                #print("\nSpark DataFrame Schema:")
                #spark_df.printSchema()
                return spark_df
                
            except Exception as e:
                print(f"Error converting to Spark DataFrame: {str(e)}")
                return None
        
        return df



from setuptools import setup, find_packages

with open("requirements.txt") as f:
	install_requires = f.read().strip().split("\n")
	
version = '1.2.8'

setup(
    name='xursparks',
    version=version,
    packages=find_packages(),
    python_requires=">=3.10",
    install_requires=install_requires,
    author='Randell Gabriel Santos',
    author_email='randellsantos@gmail.com',
    description='Encapsulating Apache Spark for Easy Usage',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/dev-doods687/xursparks',
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)
